unit AnimatedMenus98;

{$BOOLEVAL OFF}

// This conditional disable AutoBitmaps(tm) feature
{-$DEFINE NoAutoBitmaps}


interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, StdCtrls, ExtCtrls,
  Forms, Dialogs, Menus, Buttons, ComCtrls, TB97;


type
  // animation effect
  TAnimation = (anNone, anVSlide, anHSlide, anSmart, anUnfold, anPopup, anRandom);

  // button's orientation
  TButtonOrientation = (orAutomatic, orHorizontal, orVertical);

  // system type for menu
  TSystemType = (stAutomatic, stMdiFrameSystem, stMdiChildSystem);

  // menu title alignment
  TTitleAlign = (alLeft, alRight);

  // highlight style
  THighlight = (hiDefault, hiRectangle, hiBevelLowered, hiBevelRaised, hiShadow,
    hiHyperlink, hiFlash, hiShining);


  TToolbarButton98 = class;
  TTitle = class;
  TSubMenus = class;
  TColors = class;
  TOptions = class;


  // main window menu
  TMainMenu98 = class(TToolbar97)
  private
    SpaceButton      : TSpeedButton;
    Timer            : TTimer;
    FMainMenu        : TMainMenu;
    KeepSelected     : Boolean;
    Updating         : Boolean;
    FSBPanelNo       : Integer;
    FStatusBar       : TStatusBar;
    FSubMenus        : TSubMenus;
    FOnPopup         : TNotifyEvent;
    FOnMenuCommand   : TNotifyEvent;
    FOnMenuClose     : TNotifyEvent;
    FOptions         : TOptions;
    FMergedMenusList : TList;
    FGrayInactive    : Boolean;
    FSystemFont      : Boolean;

    procedure SetMainMenu(Value: TMainMenu);
    procedure SetSubMenus(Value: TSubMenus);
    procedure SetSystemFont(Value: Boolean);
    function IsFontStored: Boolean;

    procedure CheckForMDIChildMenu(Sender: TObject);
    procedure MenuSystemClick(Sender: TObject);
    procedure EndUpdate1;
    procedure EasyUpdate(MenuItem: TMenuItem);
    procedure SetMenuProperties(Menu: TPopupMenu);
    procedure MergeMenus;

    function CreateButtonForMenu(MenuItem: TMenuItem; AOwner: TComponent;
      SystemType: TSystemType): TToolbarButton98;

  protected
    procedure Loaded; override;
    procedure SetBounds(ALeft, ATop, AWidth, AHeight: Integer); override;

  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;

    procedure Rebuild;     virtual;
    procedure BeginUpdate; virtual;
    procedure EndUpdate;   virtual;

  published
    property MainMenu        : TMainMenu    read FMainMenu        write SetMainMenu;
    property StatusBar       : TStatusBar   read FStatusBar       write FStatusBar;
    property StatusBarIndex  : Integer      read FSBPanelNo       write FSBPanelNo;
    property Submenus        : TSubMenus    read FSubMenus        write SetSubMenus;
    property Options         : TOptions     read FOptions         write FOptions;
    property GrayWhenInactive: Boolean      read FGrayInactive    write FGrayInactive   default True;
    property SystemFont      : Boolean      read FSystemFont      write SetSystemFont   default False;

    property OnPopup         : TNotifyEvent read FOnPopup         write FOnPopup;
    property OnMenuCommand   : TNotifyEvent read FOnMenuCommand   write FOnMenuCommand;
    property OnMenuClose     : TNotifyEvent read FOnMenuClose     write FOnMenuClose;

    property Font stored IsFontStored;
    property ParentFont;
    property ParentShowHint;

  end;

  TOnlyMainMenu = class(TMainMenu)
  end;

  // buttons in menu bar
  TToolbarButton98 = class(TToolbarButton97)
  private
    PP               : TPoint;
    Timer            : TTimer;
    FOrientation     : TButtonOrientation;
    HBuffer, VBuffer : TBitmap;
    HBState, VBState : Boolean;
    SystemButton     : Boolean;
    FAutoSize        : Boolean;

    procedure wmLButtonDblClk(var Msg: TWMMouse);    message wm_LButtonDblClk;
    procedure cmDialogChar(var Msg: TCMDialogChar);  message cm_DialogChar;
    procedure cmMouseLeave(var Msg: TMessage);       message cm_MouseLeave;
    procedure cmMouseEnter(var Msg: TMessage);       message cm_MouseEnter;
    procedure cmSysColorChange(var Msg: TMessage);   message cm_SysColorChange;
    procedure SetOrientation(Value: TButtonOrientation);
    procedure SetAutoSize(Value: Boolean);
    procedure DelaySystemMenu(Sender: TObject);

  protected
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure MouseMove(Shift: TShiftState; X, Y: Integer); override;
    procedure Paint; override;
    procedure Click; override;

  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;

    procedure SetBounds(ALeft, ATop, AWidth, AHeight: Integer); override;
    function IsVertical: Boolean;

  published
    property Orientation: TButtonOrientation read FOrientation write SetOrientation default orAutomatic;
    property AutoSize   : Boolean            read FAutoSize    write SetAutoSize    default True;

  end;


  // popup menu
  TPopupMenu98 = class(TPopupMenu)
  private
    FSBPanelNo       : Integer;
    FStatusBar       : TStatusBar;
    MenuHandle       : THandle;
    FOnMenuCommand   : TNotifyEvent;
    FOnMenuClose     : TNotifyEvent;
    Form             : Forms.TForm;
    FOptions         : TOptions;
    FFont            : TFont;
    FParentFont      : Boolean;
    FParentShowHint  : Boolean;
    FShowHint        : Boolean;
    FSubmenus        : TSubmenus;
    FSystemFont      : Boolean;

    function IsFontStored: Boolean;
    function IsShowHintStored: Boolean;
    procedure SetFont(Value: TFont);
    procedure SetParentFont(Value: Boolean);
    procedure SetShowHint(Value: Boolean);
    procedure SetParentShowHint(Value: Boolean);
    procedure SetSubMenus(Value: TSubMenus);
    procedure SetSystemFont(Value: Boolean);

    function Button: TToolbarButton98;

  public
    MenuItems        : TMenuItem;

    constructor Create(AOwner: TComponent); override;
    destructor Destroy;                     override;

    procedure Popup(X, Y: Integer);         override;
    function FormOnScreen: Boolean;

  published
    property StatusBar       : TStatusBar   read FStatusBar       write FStatusBar;
    property StatusBarIndex  : Integer      read FSBPanelNo       write FSBPanelNo;
    property Options         : TOptions read FOptions     write FOptions;
    property Submenus        : TSubMenus    read FSubMenus        write SetSubMenus;
    property ShowHint        : Boolean      read FShowHint        write SetShowHint       stored IsShowHintStored;
    property ParentShowHint  : Boolean      read FParentShowHint  write SetParentShowHint default True;
    property Font            : TFont        read FFont            write SetFont           stored IsFontStored;
    property ParentFont      : Boolean      read FParentFont      write SetParentFont     default True;
    property SystemFont      : Boolean      read FSystemFont      write SetSystemFont     default False;

    property OnMenuCommand   : TNotifyEvent read FOnMenuCommand   write FOnMenuCommand;
    property OnMenuClose     : TNotifyEvent read FOnMenuClose     write FOnMenuClose;

  end;


  // system menu component
  TSystemMenu98 = class(TPopupMenu98)
  private
    FOnShow          : TNotifyEvent;
    FEnabled         : Boolean;
    FAllowAdjust     : Boolean;

  public
    constructor Create(AOwner: TComponent); override;

  published
    property Enabled         : Boolean      read FEnabled         write FEnabled          default True;
    property AllowAdjust     : Boolean      read FAllowAdjust     write FAllowAdjust      default True;
    property OnShow          : TNotifyEvent read FOnShow          write FOnShow;

  end;


  // menu item bitmaps (in assistance with Rolf Moser)
  TMenuItemBitmaps = class(TComponent)
  private
    FCheckedBitmap           : TBitmap;
    FUncheckedBitmap         : TBitmap;
    FMenuItem                : TMenuItem;
    FSingleBitmap            : Boolean;
    FCheckedBitmapResource   : String;
    FUncheckedBitmapResource : String;
    FEnabled                 : Boolean;

    procedure SetCheckedBitmap(Value: TBitmap);
    procedure SetUncheckedBitmap(Value: TBitmap);
    procedure SetMenuItem(Value: TMenuItem);
    procedure SetSingleBitmap(Value: Boolean);
    procedure SetCheckedBitmapResource(Value: String);
    procedure SetUncheckedBitmapResource(Value: String);
    procedure SetEnabled(Value: Boolean);

  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;

    procedure Update;

  published
    property CheckedBitmap        : TBitmap   read FCheckedBitmap           write SetCheckedBitmap;
    property UncheckedBitmap      : TBitmap   read FUncheckedBitmap         write SetUncheckedBitmap;
    property MenuItem             : TMenuItem read FMenuItem                write SetMenuItem;
    property SingleBitmap         : Boolean   read FSingleBitmap            write SetSingleBitmap            default False;
    property CheckedBmpResource   : String    read FCheckedBitmapResource   write SetCheckedBitmapResource;
    property UncheckedBmpResource : String    read FUncheckedBitmapResource write SetUncheckedBitmapResource;
    property Enabled              : Boolean   read FEnabled                 write SetEnabled                 default True;

  end;


  // menu item's overriden object for user's updates
  TMenuItem = class(Menus.TMenuItem)
  protected
    function GetBreak: TMenuBreak;
    function GetCaption: String;
    function GetChecked: Boolean;
    function GetDefault: Boolean;
    function GetEnabled: Boolean;
    function GetGroupIndex: Byte;
    function GetRadioItem: Boolean;
    function GetShortCut: TShortCut;
    function GetVisible: Boolean;
    function GetHint: String;
    function GetOnClick: TNotifyEvent;
    procedure SetBreak(Value: TMenuBreak);
    procedure SetCaption(const Value: string);
    procedure SetChecked(Value: Boolean);
    procedure SetDefault(Value: Boolean);
    procedure SetEnabled(Value: Boolean);
    procedure SetGroupIndex(Value: Byte);
    procedure SetRadioItem(Value: Boolean);
    procedure SetShortCut(Value: TShortCut);
    procedure SetVisible(Value: Boolean);
    procedure SetHint(Value: String);
    procedure SetOnClick(Value: TNotifyEvent);

  public
    procedure Insert(Index: Integer; Item: Menus.TMenuItem);
    procedure Delete(Index: Integer);
    procedure Add(Item: Menus.TMenuItem);
    procedure Remove(Item: Menus.TMenuItem);

    function IsFirstLevelItem: Boolean;

  published
    property Break      : TMenuBreak   read GetBreak      write SetBreak      default mbNone;
    property Caption    : String       read GetCaption    write SetCaption;
    property Checked    : Boolean      read GetChecked    write SetChecked    default False;
    property Default    : Boolean      read GetDefault    write SetDefault    default False;
    property Enabled    : Boolean      read GetEnabled    write SetEnabled    default True;
    property GroupIndex : Byte         read GetGroupIndex write SetGroupIndex default 0;
    property Hint       : string       read GetHint       write SetHint;
    property RadioItem  : Boolean      read GetRadioItem  write SetRadioItem  default False;
    property ShortCut   : TShortCut    read GetShortCut   write SetShortCut   default 0;
    property Visible    : Boolean      read GetVisible    write SetVisible    default True;
    property OnClick    : TNotifyEvent read GetOnClick    write SetOnClick;

  end;


  // Form with SystemMenu98
  TForm = class(Forms.TForm)
  private
    PP               : TPoint;
    Timer            : TTimer;
    FOldMenu         : TMainMenu;
    ADoAdjust        : Boolean;

    procedure wmNCLButtonDown  (var Msg: TWMNCLButtonDown);   message wm_NCLButtonDown;
    procedure wmNCLButtonDblClk(var Msg: TWMNCLButtonDblClk); message wm_NCLButtonDblClk;
    procedure wmNCLButtonUp    (var Msg: TWMNCLButtonUp);     message wm_NCLButtonUp;
    procedure wmNCRButtonDown  (var Msg: TWMNCRButtonDown);   message wm_NCRButtonDown;
    procedure wmNCRButtonUp    (var Msg: TWMNCRButtonUp);     message wm_NCRButtonUp;

    procedure SystemClick(Sender: TObject);

  protected
    SystemMenu98     : TSystemMenu98;

    procedure Loaded; override;
    procedure DelayPopupSysMenu(Sender: TObject); virtual;

  public
    property Menu    : TMainMenu read FOldMenu;

    procedure PopupSysMenu(X, Y: Integer; DoAdjust, DoDelay: Boolean); virtual;

  end;


  // menu title
  TTitle = class(TPersistent)
  private
    FAlign           : TTitleAlign;
    FAlignment       : TAlignment;
    FText            : String;
    FFont            : TFont;
    FColorBegin      : TColor;
    FColorEnd        : TColor;
    FWidth           : Integer;
    FDirection       : Boolean;
    FVisible         : Boolean;
    FBitmap          : TBitmap;

    procedure SetFont(Value: TFont);
    procedure SetBitmap(Value: TBitmap);
    function IsBitmapStored: Boolean;

  public
    constructor Create;
    destructor Destroy; override;
    procedure Paint(Canvas: TCanvas);
    function TextAlign(Width, TextWidth: Integer): Integer;
    procedure Assign(Source: TPersistent); override;

  published
    property Align      : TTitleAlign read FAlign      write FAlign      default alLeft;
    property Alignment  : TAlignment  read FAlignment  write FAlignment  default taLeftJustify;
    property Text       : String      read FText       write FText;
    property Font       : TFont       read FFont       write SetFont;
    property ColorBegin : TColor      read FColorBegin write FColorBegin default clBlack;
    property ColorEnd   : TColor      read FColorEnd   write FColorEnd   default clNavy;
    property Width      : Integer     read FWidth      write FWidth      default 50;
    property Direction  : Boolean     read FDirection  write FDirection  default True;
    property Visible    : Boolean     read FVisible    write FVisible    default False;
    property Bitmap     : TBitmap     read FBitmap     write SetBitmap   stored IsBitmapStored;

  end;


  // menu colors
  TColors = class(TPersistent)
  private
    FBorder          : TColor;
    FMenu            : TColor;
    FMenuText        : TColor;
    FMenuShadow      : TColor;
    FHighlight       : TColor;
    FHighlightText   : TColor;
    FDisabledText    : TColor;
    FDisabledShadow  : TColor;
    FLine            : TColor;
    FLineShadow      : TColor;

  public
    constructor Create;
    function IsDefaultColors: Boolean;
    procedure Assign(Source: TPersistent); override;

  published
    property Border         : TColor   read FBorder         write FBorder         default clNone;
    property Menu           : TColor   read FMenu           write FMenu           default clMenu;
    property MenuText       : TColor   read FMenuText       write FMenuText       default clMenuText;
    property MenuShadow     : TColor   read FMenuShadow     write FMenuShadow     default clBtnShadow;
    property Highlight      : TColor   read FHighlight      write FHighlight      default clHighlight;
    property HighlightText  : TColor   read FHighlightText  write FHighlightText  default clHighlightText;
    property DisabledText   : TColor   read FDisabledText   write FDisabledText   default clBtnShadow;
    property DisabledShadow : TColor   read FDisabledShadow write FDisabledShadow default clBtnHighlight;
    property Line           : TColor   read FLine           write FLine           default clBtnShadow;
    property LineShadow     : TColor   read FLineShadow     write FLineShadow     default clBtnHighlight;

  end;


  // menu options
  TBaseOptions = class(TCollectionItem)
  private
    FItemHeight      : Integer;
    FSeparatorHeight : Integer;
    FIndent          : Integer;
    FTextAlignment   : TAlignment;
    FTitle           : TTitle;
    FBackground      : TBitmap;
    FHighlight       : THighlight;
    FColors          : TColors;
    FCtl3D           : Boolean;
    FFitToBackground : Boolean;
    FItemSpace       : Integer;
    FTireAwayable    : Boolean;

    function IsBackgroundStored: Boolean;
    function IsColorsStored: Boolean;
    procedure SetBackground(Value: TBitmap);
    procedure SetColors(Value: TColors);
    procedure SetTitle(Value: TTitle);

  public
    constructor Create(Collection: TCollection); override;
    destructor Destroy; override;
    procedure Assign(Source: TPersistent); override;

  published
    property ItemHeight      : Integer      read FItemHeight      write FItemHeight       default 19;
    property SeparatorHeight : Integer      read FSeparatorHeight write FSeparatorHeight  default 11;
    property Indent          : Integer      read FIndent          write FIndent           default 27;
    property Alignment       : TAlignment   read FTextAlignment   write FTextAlignment    default taLeftJustify;
    property Title           : TTitle       read FTitle           write SetTitle;
    property Background      : TBitmap      read FBackground      write SetBackground     stored  IsBackgroundStored;
    property Highlight       : THighlight   read FHighlight       write FHighlight        default hiDefault;
    property Colors          : TColors      read FColors          write SetColors         stored  IsColorsStored;
    property Ctl3D           : Boolean      read FCtl3D           write FCtl3D            default True;
    property FitToBackground : Boolean      read FFitToBackground write FFitToBackground  default False;
    property ItemSpace       : Integer      read FItemSpace       write FItemSpace        default 0;
    property TireAwayable    : Boolean      read FTireAwayable    write FTireAwayable     default False;

  end;

  TOptions = class(TBaseOptions)
  private
    FOpenAnimation   : TAnimation;
    FCloseAnimation  : TAnimation;
    FTimeout         : Integer;
    FDoubleSpeed     : Boolean;

  public
    constructor Create(Collection: TCollection); override;
    procedure Assign(Source: TPersistent); override;

  published
    property OpenAnimation   : TAnimation   read FOpenAnimation   write FOpenAnimation    default anSmart;
    property CloseAnimation  : TAnimation   read FCloseAnimation  write FCloseAnimation   default anSmart;
    property Timeout         : Integer      read FTimeout         write FTimeout          default 20;
    property DoubleSpeed     : Boolean      read FDoubleSpeed     write FDoubleSpeed     default False;

  end;

  TSubmenuOptions = class(TBaseOptions)
  private
    FFont            : TFont;
    FParentFont      : Boolean;
    FShowHint        : Boolean;
    FMenuItem        : Menus.TMenuItem;

    function IsFontStored: Boolean;
    procedure SetFont(Value: TFont);
    procedure SetParentFont(Value: Boolean);
    procedure SetMenuItem(Value: Menus.TMenuItem);

{$IFDEF VER100}
    function GetDisplayName: string; override;
{$ENDIF}

  public
    constructor Create(Collection: TCollection); override;
    destructor Destroy; override;
    procedure Assign(Source: TPersistent); override;

  published
    property ShowHint        : Boolean      read FShowHint        write FShowHint         default False;
    property Font            : TFont        read FFont            write SetFont           stored IsFontStored;
    property ParentFont      : Boolean      read FParentFont      write SetParentFont     default True;
    property MenuItem        : Menus.TMenuItem read FMenuItem     write SetMenuItem;

  end;

  
  TSubmenus = class(TCollection)
  private
    FPersistent: TPersistent;
    function GetItem(Index: Integer): TSubmenuOptions;
    procedure SetItem(Index: Integer; Value: TSubmenuOptions);

  protected
{$IFDEF VER100}
    function GetOwner: TPersistent; override;
{$ENDIF}
    procedure Update(Item: TCollectionItem); override;

  public
    constructor Create(Persistent: TPersistent);
    function Add: TSubmenuOptions;
    property Items[Index: Integer]: TSubmenuOptions read GetItem write SetItem; default;

  end;


const
  // Custom Sounds - set your favorite
  MenuPopupSound    : String = 'MenuPopup';      // or = 'c:\laser.wav';
  MenuCloseSound    : String = '';
  MenuCommandSound  : String = 'MenuCommand';

  // Keyboard shortcuts
  KeyboardShortCuts : TStringList = nil;


function NewSubMenu(const ACaption: string; hCtx: Word; const AName: string;
  Items: array of Menus.TMenuItem): TMenuItem;
function NewItem(const ACaption: string; AShortCut: TShortCut;
  AChecked, AEnabled: Boolean; AOnClick: TNotifyEvent; hCtx: Word;
  const AName: string): TMenuItem;
function NewLine: TMenuItem;

procedure Register;

function LoadBitmapByName(S: String): HBitmap;


implementation
uses
  CommCtrl, MmSystem, ShellApi;


{$R PopupMenu98Form.Dfm }
{$R AnimatedMenus98.res }

{$IFNDEF NoAutoBitmaps}
{$R AutoBitmaps.res}
{$ENDIF}


type
  THintWindow98 = class(THintWindow)
  private
    Timer: TTimer;
    Hint: TStringList;

    procedure OnTimer(Sender: TObject);

  protected
    procedure Paint; override;

  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure ActivateHint(Rect: TRect; const AHint: string); override;
    procedure ShowHint(const AHint: string);

  end;


  TPopupMenu98Form = class(Forms.TForm)
  private
//    L: TList;

    Buffer: TBitmap;
    FSelectedIndex, LastSelectedIndex, MenuItemCount, MainWidth, DX, DY: Integer;
    OpenAnimation, CloseAnimation: TAnimation;
    SelectedChanged, PaintMenu, FromBottomToTop, FromRightToLeft, DrawDisabled,
    IgnoreFirstMouseMove: Boolean;
    PopupMenu: TPopupMenu98;
    CurSubMenu, ParentPopupMenu98Form: TPopupMenu98Form;
    Timer: TTimer;
    MenuHandle: THandle;
    Options: TBaseOptions;

    procedure SetSelectedIndex(Value: Integer);
    procedure PopupMenu98FormTimer(Sender: TObject);

    procedure CMMouseLeave(var Msg: TMessage);     message cm_MouseLeave;
    procedure WMSetFocus(var Msg: TWMSetFocus);    message wm_SetFocus;
    procedure WMActivate(var Msg: TWMActivate);    message wm_Activate;
//    procedure WMActivate(var Msg: TWMActivate);    message wm_Activate;
    procedure WMMouseActivate(var Msg: TWMMouseActivate); message wm_MouseActivate;
    procedure CMShowingChanged(var Msg: TMessage); message cm_ShowingChanged;

  protected
    function GetRealHeight: Integer; virtual;
    function GetRealWidth: Integer;  virtual;
    function CurItemTop: Integer;    virtual;

    procedure Paint; override;
//    procedure PaintWindow(DC: HDC); override;
    procedure MouseMove(Shift: TShiftState; X, Y: Integer); override;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure SetZOrder(TopMost : Boolean); override;

  public
    CurMenuItem, MenuItems: Menus.TMenuItem;
    Animated: Boolean;
    KillAnimation: Boolean;

    property SelectedIndex: Integer read FSelectedIndex write SetSelectedIndex;

    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function GetIndexAt(X, Y: Integer): Integer;

    function Animate: Boolean; virtual;
    function AnimatedHide: Boolean; virtual;
    procedure SilentShow;
    procedure SilentHide;

    procedure PopupCurSubMenu;
    procedure BringMenuToFront;
    function GetCurMenuItem(RaiseException: Boolean): Menus.TMenuItem;
    function GetMenuItemIndex(Index: Integer; RaiseException: Boolean): Menus.TMenuItem;

  end;


const
  // directions for GetNextToolbarButton98
  drLeft  = -1;
  drRight = 1;

  vk_0 = Byte('0');
  vk_Z = Byte('Z');

  FormFlags = swp_NoMove or swp_NoSize or swp_NoActivate;

const
  ActivePopupMenu98       : TPopupMenu98     = nil;
  ActiveMenu98List        : TList            = nil;
  ActiveHintWindow98      : THintWindow98    = nil;
  Pattern                 : TBitmap          = nil;
  IgnoreNextMenuUp        : Boolean          = False; // ignore alt key up after alt key down
  CheckMarkBitmap         : HBitmap          = 0;
  RadioItemBitmap         : HBitmap          = 0;
  HGetMessageHook         : THandle          = 0;
  LastButtonMouseInControl: TToolbarButton97 = nil;
  DesktopWindow           : THandle          = 0;
  DesktopLocksCount       : Integer          = 0;

var
  Z: array [0..256] of Char;
  mii: TMenuItemInfo;


{ Routines }

function AssignedActivePopupMenu98Form: Boolean;
begin
  Result:= Assigned(ActivePopupMenu98) and ActivePopupMenu98.FormOnScreen;
end;


function ActiveMenu98: TMainMenu98;
// check for current active MainMenu98
var
  GAW: HWND;
  I: Integer;
begin
  Result:= nil;
  GAW:= GetActiveWindow;

  for I:= 0 to ActiveMenu98List.Count -1 do
    if (Result = nil)
    or (Forms.TForm(TMainMenu98(ActiveMenu98List[I]).Owner).Handle = GAW)
    then Result:= TMainMenu98(ActiveMenu98List[I]);
end;


function AssignedActiveMenu98: Boolean;
begin
  Result:= ActiveMenu98 <> nil;
end;


function GetMyActiveMenu98(MenuItem: Menus.TMenuItem): TMainMenu98;
// if MenuItem is an up-level menu item in TMainMenu,
// returns ActiveMenu98 for this TMainMenu
var
  I, J, K: Integer;
begin
  Result:= nil;

  for I:= 0 to ActiveMenu98List.Count -1 do
    with TMainMenu98(ActiveMenu98List[I]) do begin
      if Assigned(MainMenu) then
        for J:= 0 to MainMenu.Items.Count -1 do
          if MainMenu.Items[J] = MenuItem then begin
            Result:= TMainMenu98(ActiveMenu98List[I]);
            Exit;
          end;

      for K:= 0 to FMergedMenusList.Count -1 do
        with TMainMenu(FMergedMenusList[K]) do
          for J:= 0 to Items.Count -1 do
            if Items[J] = MenuItem then begin
              Result:= TMainMenu98(ActiveMenu98List[I]);
              Exit;
            end;
    end;
end;


procedure HideWindowMenu(Owner: TComponent);
begin
  // hides only MDIForm's menu
  if (Owner is Forms.TForm)
  and (not (Owner.Owner is Forms.TForm))
  and (Forms.TForm(Owner).Menu <> nil)
  then begin
    if Owner is AnimatedMenus98.TForm then
      TForm(Owner).FOldMenu:= Forms.TForm(Owner).Menu;

    Forms.TForm(Owner).Menu:= nil;
  end;
end;


procedure SetStatusBarText(HintText: String);
var
  I: Integer;
  S: String;
begin
  S:= Trim(GetLongHint(HintText));

  // remove '&#13;' symbols from status bar text
  repeat
    I:= Pos('&#', S);
    if I = 0 then System.Break;
    System.Delete(S, I, 5);
    System.Insert(' ', S, I);
  until False;

  if Assigned(ActivePopupMenu98)
  and Assigned(ActivePopupMenu98.StatusBar)
  then
    with ActivePopupMenu98, StatusBar do begin
      if SimplePanel
      then SimpleText:= S
      else
        if StatusBarIndex < Panels.Count
        then Panels[StatusBarIndex].Text:= S;

      Exit;
    end;

  if AssignedActiveMenu98
  and Assigned(ActiveMenu98.StatusBar)
  then
    with ActiveMenu98, StatusBar do
      if SimplePanel
      then SimpleText:= S
      else
        if StatusBarIndex < Panels.Count
        then Panels[StatusBarIndex].Text:= S;

end;

// processing mousemove and paint messages --
// a bit faster than Application.ProcessMessages
// thanks to Jordan Russell
procedure ProcessPaintMessages;
var
  Msg: TMsg;
begin
  while PeekMessage(Msg, 0, wm_Paint, wm_Paint, pm_NoRemove) do begin
    case Integer(GetMessage(Msg, 0, wm_Paint, wm_Paint)) of
      -1: Exit;
      0: begin PostQuitMessage(Msg.WParam); Exit; end;
    end;
    DispatchMessage(Msg);
  end;
end;

procedure ProcessMouseMoveMessages;
var
  Msg: TMsg;
begin
  while PeekMessage(Msg, 0, wm_MouseMove, wm_MouseMove, pm_NoRemove) do begin
    case Integer(GetMessage(Msg, 0, wm_MouseMove, wm_MouseMove)) of
      -1: Exit;
      0: begin PostQuitMessage(Msg.WParam); Exit; end;
    end;
    DispatchMessage(Msg);
  end;
end;


procedure SetActivePopupMenu98Up(MultimediaAction: Boolean);
var
  F, F1: TPopupMenu98Form;
begin
  if Assigned(ActivePopupMenu98)
  and Assigned(ActivePopupMenu98.PopupComponent)
  and (ActivePopupMenu98.PopupComponent is TToolbarButton98) then begin

    // down to the last submenu
    F:= ActivePopupMenu98.Form as TPopupMenu98Form;
    while Assigned(F) and Assigned(F.CurSubMenu) do
      F:= F.CurSubMenu;

    // up to the first level menu
    while Assigned(F) do begin
      if MultimediaAction
      then F.AnimatedHide
      else F.Height:= 0;

      F1:= F;
      F:= F.ParentPopupMenu98Form;
      if Assigned(F) then F1.Release;
      
    end;

    if Assigned(ActivePopupMenu98.OnMenuClose)
    then ActivePopupMenu98.OnMenuClose(nil);

    if ActivePopupMenu98.Button <> nil then
      with ActivePopupMenu98.Button do begin
        MouseLeft;
        Down:= False;
        Update;
      end;
  end;
end;


procedure KillActivePopupMenu98(MultimediaAction: Boolean);
begin
  try
    if AssignedActivePopupMenu98Form then
      with ActivePopupMenu98, Form as TPopupMenu98Form do begin
        SetStatusBarText('');
        Timer.Enabled:= False;
        SetActivePopupMenu98Up(MultimediaAction);

        if not (Assigned(ActivePopupMenu98)) then Exit;

        KillAnimation:= True;
        SilentHide;

        LastButtonMouseInControl:= Button;
      end;
  except
  end;

  ActivePopupMenu98:= nil;
end;


procedure SetFirstIndexSelected(Button: TToolbarButton97);
begin
  if Assigned(Button.DropdownMenu)
  and (Button.DropdownMenu is TPopupMenu98) then
    with TPopupMenu98Form(TPopupMenu98(Button.DropdownMenu).Form) do begin
      SelectedIndex:= 0;
      IgnoreFirstMouseMove:= True;
      DrawDisabled:= True;
    end;
end;


function GetNextToolbarButton98(Control: TControl; Direction: Integer; DoClick: Boolean): TToolbarButton98;
var
  I, J: Integer;
  AM98: TMainMenu98;
begin
  Result:= nil;
  if (not Assigned(Control)) or (not AssignedActiveMenu98) then Exit;

  I:= 0;
  AM98:= ActiveMenu98;
  for J:= 0 to ActiveMenu98List.Count -1 do
    if Control.Parent = ActiveMenu98List[J] then begin
      AM98:= TMainMenu98(ActiveMenu98List[J]);

      I:= 0;
      while (I < AM98.ControlCount)
      and (AM98.Controls[I] <> Control)
      do Inc(I);

      Break;
    end;

  repeat
    Inc(I, Direction);
    if I <= -1 then I:= AM98.ControlCount -1;
    if I >= AM98.ControlCount then I:= 0;
  until (AM98.Controls[I] is TToolbarButton98)
    and (Assigned(TToolbarButton98(AM98.Controls[I]).DropdownMenu) or (not DoClick))
    and AM98.Controls[I].Enabled;

  Result:= TToolbarButton98(AM98.Controls[I]);

  if DoClick
  and (Result <> Control)
  and Assigned(Result)
  and Assigned(Result.DropdownMenu)
  and (Result.DropdownMenu is TPopupMenu98) then begin
    SetFirstIndexSelected(Result);
    Result.Click;
  end;
end;


function SearchForActiveMenu98Key(Key: Integer; SetSelected: Boolean): TToolbarButton98;
var
  I: Integer;
begin
  Result:= nil;
  if not AssignedActiveMenu98 then Exit;

  with ActiveMenu98 do
     for I:= 0 to ControlCount -1 do
       if (Controls[I] is TToolbarButton98)
       and TToolbarButton98(Controls[I]).Enabled
       and (TToolbarButton98(Controls[I]).Caption <> '-')
       and IsAccel(Key, TToolbarButton98(Controls[I]).Caption)
       then begin
         Result:= Controls[I] as TToolbarButton98;
         SetFirstIndexSelected(Result);
         Exit;
       end;
end;


function GetMessageHook(Code, wParam, lParam: Integer): Integer; stdcall;
type
  TClickResult = (crDisabled, crClicked, crShortCutMoved);
const
  AltMask = $20000000;
var
  AMFA: Boolean; // active menu98's form is active
  AMFAH: Boolean;  // menu form is enabled
  AMFCWA: Boolean; // active menu98's form's ActiveMdiChild is assigned

  AM98: TMainMenu98;
  ActiveWindow, ActiveMdi: HWnd;
  P: TPoint;
  I, Msg, Key: Integer;
  C: TControl;
  F: TPopupMenu98Form;
  B: TToolbarButton97;
  ShortCutItem: Menus.TMenuItem;
  ShortCut: TShortCut;
  M: TMsg;

  procedure ClearMessage;
  begin
    FillChar(PMsg(lParam)^, SizeOf(TMsg), 0);
  end;

  function GetWMKeyMsg: TWMKey;
  begin
    Result.Msg:=      Msg;
    Result.CharCode:= Key;
    Result.KeyData:=  PMsg(lParam)^.lParam;
  end;

  // thanks to Borland (Inspire) for this code
  function DoClick(Item: Menus.TMenuItem): TClickResult;
  begin
    Result:= crClicked;
    if Item.Parent <> nil then Result:= DoClick(Item.Parent);
    if Result = crClicked then
      if Item.Enabled and Item.Visible then
        try
          Item.Click;
          if ShortCutItem.ShortCut <> ShortCut then
            Result:= crShortCutMoved;
        except
        end
      else Result:= crDisabled;
  end;

  function IsShortCut(MainMenu: TMainMenu): Boolean;
  var
    ClickResult: TClickResult;
  begin
    Result:= False;
    if not Assigned(MainMenu) then Exit;

    repeat
      ClickResult:= crDisabled;
      ShortCutItem:= MainMenu.FindItem(ShortCut, fkShortCut);

      if ShortCutItem <> nil then begin
        KillActivePopupMenu98(True);
        ClickResult := DoClick(ShortCutItem);
      end;

    until ClickResult <> crShortCutMoved;
    Result:= ShortCutItem <> nil;
  end;

  function IsKeyboardShortcut: Boolean;
  var
    I: Integer;
    S: String;
    SC: TShortCut;
  begin
    SC:= ShortCut;
    Result:= False;
    if not Assigned(KeyboardShortCuts) then Exit;

    try
      ShortCutItem:= nil;
      S:= ShortCutToText(ShortCut);

      for I:= 0 to KeyboardShortCuts.Count -1 do
        if S = KeyboardShortCuts[I]
        then ShortCutItem:= TMenuItem(KeyboardShortCuts.Objects[I]);

      if ShortCutItem <> nil then begin
        KillActivePopupMenu98(True);
        ShortCut:= ShortCutItem.ShortCut;
        DoClick(ShortCutItem);
      end;

    except
      ShortCutItem:= nil;

    end;

    Result:= ShortCutItem <> nil;
    ShortCut:= SC;
  end;

  function IsMenuShortCut: Boolean;
  var
    I: Integer;
  begin
    ShortCut:= Key;
    if GetKeyState(vk_Shift) < 0 then Inc(ShortCut, scShift);
    if GetKeyState(vk_Control) < 0 then Inc(ShortCut, scCtrl);
    if PMsg(lParam)^.lParam and AltMask <> 0 then Inc(ShortCut, scAlt);

    Result:= Assigned(AM98) and (IsShortCut(AM98.MainMenu) or IsKeyboardShortcut);

    I:= 0;
    while (not Result)
      and Assigned(AM98)
      and Assigned(AM98.FMergedMenusList)
      and (I < AM98.FMergedMenusList.Count)
    do
      Result:= IsShortCut(TMainMenu(AM98.FMergedMenusList));
  end;

  procedure SearchForOpenedMenuShortcut;
  var
    I: Integer;
  begin
    // just a symbol key - seeking for menu item
    for I:= 0 to F.MenuItems.Count -1 do
      if F.MenuItems[I].Visible
      and IsAccel(Key, F.MenuItems[I].Caption) then
        with F.MenuItems[I] do begin
          if (not Enabled) then System.Break;

          if Count = 0 then try // ordinary item
            KillActivePopupMenu98(True);
            sndPlaySound(PChar(MenuCommandSound), snd_Async + snd_NoDefault);
            if Assigned(OnClick) then OnClick(F.MenuItems[I]);
          except
          end
          else begin // SubMenu
            F.SelectedIndex:= I;
            F.SelectedChanged:= False;
            F.PopupCurSubMenu;
          end;

          System.Break;
        end { for } ;

    if I >= F.MenuItems.Count then
      MessageBeep(-1);
  end;

  function IsOkControl(C: TControl): Boolean;
  begin
    Result:= False;
    if Assigned(C) then Result:= IsOkControl(C.Parent);
    Result:= Result or (C is TPopupMenu98Form)
      or ((C = ActivePopupMenu98.PopupComponent) and (C is TToolbarButton97));
  end;

begin
  Result:= 0;
  ActiveWindow:= GetActiveWindow; // a bit of optimization

  if (Code >= 0)
  and Assigned(Application)
  and Application.Active
  and (not IsIconic(ActiveWindow))
  then begin
    AM98:= ActiveMenu98;
    Msg:= PMsg(lParam)^.Message;
    Key:= PMsg(lParam)^.wParam;

    // check for mouse messages
    if ((Msg >= wm_LButtonDblClk) and (Msg <= wm_MButtonDblClk))
    or ((Msg >= wm_NCRButtonDblClk) and (Msg <= wm_NCMButtonDblClk))
    or (Msg = wm_LButtonDown)
    or (Msg = wm_NCLButtonDown)
    or (Msg = wm_NCRButtonDown)
    then
      if AssignedActivePopupMenu98Form then begin
        GetCursorPos(P);
        C:= FindDragTarget(P, True);
        if IsOkControl(C) then Exit;

        KillActivePopupMenu98(True);
        IgnoreNextMenuUp:= False;
        ShowCaret(0);
      end
      else
        if Assigned(AM98) and Assigned(ButtonMouseInControl) then begin
          ButtonMouseInControl.MouseLeft;
          AM98.KeepSelected:= False;
          ShowCaret(0);
        end;


    // check for system menu calls
    if (Msg = wm_SysKeyDown)
    and (Key = vk_Space)
    then begin
      PostMessage(ActiveWindow, wm_NCLButtonDown, htSysMenu, 0);
      ClearMessage;
    end;

    // another key?
    if ((Msg = wm_KeyDown) or (Msg = wm_KeyUp) or (Msg = wm_SysKeyDown) or (Msg = wm_SysKeyUp))
    then begin

      // initialization
      AMFA:= False;
      AMFAH:= False;
      AMFCWA:= False;
      ActiveMdi:= 0;
      if Assigned(AM98) then
      with AM98.Owner as Forms.TForm do begin
        AMFA:= (ActiveWindow = Handle)
          or ((FormStyle = fsNormal) and (Application.MainForm = AM98.Owner));
        AMFAH:= IsWindowEnabled(Handle) and Enabled;

        if Assigned(ActiveMdiChild) then
          ActiveMdi:= ActiveMdiChild.Handle;

        AMFCWA:= ActiveMdi = ActiveWindow;
      end;


      // shortcuts
      if ((Msg = wm_KeyDown) or (Msg = wm_SysKeyDown))
      and Assigned(AM98)
      and AMFAH
      and (AMFA or (ActiveWindow = ActiveMdi))
      and IsMenuShortCut
      then begin
        ClearMessage;
        Exit;
      end;


      // getting the topmost MenuForm
      F:= nil;
      if Assigned(ActivePopupMenu98) then begin
        F:= ActivePopupMenu98.Form as TPopupMenu98Form;
        while F.CurSubMenu <> nil do F:= F.CurSubMenu;
      end;


      // system menu for child window
      if (Msg = wm_SysKeyDown)
      and ((Key = $BD) or (Key = $6D))
      and AMFCWA
      then begin
        PostMessage(ActiveMdi, wm_NCLButtonDown, htSysMenu, 0);
        ClearMessage;
      end;


      // check for vk_Menu and vk_F10 - key down
      if(((Msg = wm_SysKeyDown) and ((Key = vk_Menu) or (Key = vk_F10)))
      or ((Msg = wm_KeyDown) and (Key = vk_F10)))
      and AMFA
      and AMFAH
      and (Assigned(ButtonMouseInControl) or AssignedActivePopupMenu98Form)
      then begin
        KillActivePopupMenu98(True);

        if Assigned(LastButtonMouseInControl) then begin
          LastButtonMouseInControl.MouseLeft;
          LastButtonMouseInControl.Down:= False;
        end;

        if Assigned(AM98) then
          Forms.TForm(AM98.Owner).SetFocus;

        ShowCaret(0);

        if Assigned(AM98)
        then AM98.KeepSelected:= False;

        IgnoreNextMenuUp:= True;
        ClearMessage;
      end { wm_SysKeyDown };


      // check for vk_Menu and vk_F10 - key up
      if(((Msg = wm_SysKeyUp) and ((Key = vk_Menu) or (Key = vk_F10)))
      or ((Msg = wm_KeyUp) and (Key = vk_F10)))
      and AMFA
      and AMFAH
      and Assigned(AM98)
      then begin
        if not Assigned(LastButtonMouseInControl)
        then begin
          I:= 0;
          while (I < AM98.ControlCount)
          and ((not (AM98.Controls[I] is TToolbarButton98))
          or TToolbarButton98(AM98.Controls[I]).SystemButton)
          do Inc(I);

          if (I < 0) or (I >= AM98.ControlCount) then Exit;
          LastButtonMouseInControl:= TToolbarButton98(AM98.Controls[I])
        end;

        if (TToolbarButton98(LastButtonMouseInControl).FState <> bsMouseIn)
        and (not IgnoreNextMenuUp) then begin
          ButtonMouseInControl:= LastButtonMouseInControl;
          ButtonMouseInControl.MouseEntered;
          AM98.KeepSelected:= True;
          HideCaret(0);
        end;

        IgnoreNextMenuUp:= False;
        ClearMessage;
      end { wm_SysKeyUp };



      // check for shortcuts on main menu
      // thanks to Jordan Russell for remark
      if  (Msg = wm_SysKeyDown)
      and (Key >= vk_0) and (Key <= vk_Z)
      and (AMFA or AMFCWA)
      and AMFAH
      then
        if AssignedActivePopupMenu98Form
        then
          SearchForOpenedMenuShortcut
        else begin
          LastButtonMouseInControl:= SearchForActiveMenu98Key(Key, True);

          if Assigned(LastButtonMouseInControl) then begin
            if Assigned(ButtonMouseInControl) then
              ButtonMouseInControl.MouseLeft;

            ButtonMouseInControl:= nil;
            LastButtonMouseInControl.Click;
            LastButtonMouseInControl.MouseLeft;

            ShowCaret(0);
            AM98.KeepSelected:= False;
            IgnoreNextMenuUp:= True;
            ClearMessage;
          end;
        end { wm_SysKeyDown };


      // check for other keys
      if (Msg = wm_KeyDown) then begin
        IgnoreNextMenuUp:= False;

        // -------------------
        // main menu is active
        // ActiveMenu98.KeepSelected (i.e. 'selected by kbd')
        // is necessary - thanks to Ulf Haueisen
        if Assigned(AM98)
        and AMFAH
        and AM98.KeepSelected
        and Assigned(ButtonMouseInControl)
        then begin
          case Key of

            vk_Escape: begin
              ShowCaret(0);
              AM98.KeepSelected:= False;
              LastButtonMouseInControl:= ButtonMouseInControl;
              ButtonMouseInControl.MouseLeft;
            end { vk_Escape } ;

            vk_Left: begin
              LastButtonMouseInControl:= ButtonMouseInControl;
              ButtonMouseInControl.MouseLeft;
              ButtonMouseInControl:= GetNexTToolbarButton98(LastButtonMouseInControl, drLeft, False);
              ButtonMouseInControl.MouseEntered;
            end { vk_Left } ;

            vk_Right: begin
              LastButtonMouseInControl:= ButtonMouseInControl;
              ButtonMouseInControl.MouseLeft;
              ButtonMouseInControl:= GetNexTToolbarButton98(LastButtonMouseInControl, drRight, False);
              ButtonMouseInControl.MouseEntered;
            end { vk_Right } ;

            vk_Return, vk_Down, vk_Up, vk_Space: begin
              AM98.KeepSelected:= False;
              SetFirstIndexSelected(ButtonMouseInControl);
              ButtonMouseInControl.Click;
            end { vk_Return } ;

            else begin
              // alphabetical keys - search for key and beep if not found
              // thank Jan Hulala for notifing about this bug
                B:= SearchForActiveMenu98Key(Key, True);

                if Assigned(B)
                then begin
                  ButtonMouseInControl.MouseLeft;
                  B.Click;
                end
                else MessageBeep(-1);
              end;
          end { case } ;

          ClearMessage;
        end { Assigned(ButtonMouseInControl) } ;



        // -------------------
        // menu form is opened
        if Assigned(F) then begin
          case Key of

            vk_Return:
              F.MouseUp(mbLeft, [], 1, 1);

            vk_Escape:
              if not Assigned(F.ParentPopupMenu98Form)
              then begin
                KillActivePopupMenu98(True);
                ButtonMouseInControl:= LastButtonMouseInControl;
                if Assigned(LastButtonMouseInControl) then
                  LastButtonMouseInControl.MouseEntered;

                if AssignedActiveMenu98 then
                  ActiveMenu98.KeepSelected:= True;
                HideCaret(0);
              end
              else begin
                F.ParentPopupMenu98Form.CurSubMenu:= nil;
                F.AnimatedHide;
              end { vk_Escape } ;

            vk_Left:
              if not Assigned(F.ParentPopupMenu98Form)
              then
                // I like it! Thanks to Favio E. Ugalde
                if ActivePopupMenu98.Button <> nil
                then GetNextToolbarButton98(ActivePopupMenu98.Button, drLeft, True)
                else F.SelectedIndex:= 0

              else begin
                F.ParentPopupMenu98Form.CurSubMenu:= nil;
                F.AnimatedHide;
              end { vk_Left } ;

            vk_Up: begin
              F.SelectedIndex:= F.SelectedIndex -1;
              F.Timer.Enabled:= False;
            end { vk_Up } ;

            vk_Right:
              if Assigned(F.GetCurMenuItem(False))
              and (F.CurMenuItem.Count > 0)
              then F.PopupCurSubMenu
              else
                // I like it too! Thanks to Favio E. Ugalde again
                if ActivePopupMenu98.Button <> nil
                then GetNexTToolbarButton98(ActivePopupMenu98.Button, drRight, True)
                else F.SelectedIndex:= -1;

            vk_Down: begin
              F.SelectedIndex:= F.SelectedIndex +1;
              F.Timer.Enabled:= False;
            end { vk_Down } ;

            vk_F1:
              if (Application.HelpFile <> '')
              and Assigned(F.GetCurMenuItem(False))
              and (F.CurMenuItem.HelpContext <> 0)
              then begin
                // this piece of code programmed by Victor Santos
                I := F.CurMenuItem.HelpContext;
                KillActivePopupMenu98(True);

                if Assigned(AM98)
                and (biHelp in Forms.TForm(AM98.Owner).BorderIcons)
                then WinHelp(ActiveWindow, PChar(Application.HelpFile), Help_ContextPopup, I)
                else WinHelp(ActiveWindow, PChar(Application.HelpFile), Help_Context, I);

                // We have to remove the next message that is in the queue.
                PeekMessage(M, 0, 0, 0, pm_Remove);
              end;

            else // alphabetical keys - search for shortcut
              if (Key <> vk_Shift) and (Key <> vk_Control) then
                SearchForOpenedMenuShortcut;

          end { case } ;

          ClearMessage;
        end { Assigned(F) } ;
      end { another key } ;
    end { keyboard message }
  end { main form is active }
  else { code <> hs_Action }
    Result:= CallNextHookEx(HGetMessageHook, Code, wParam, lParam);
end;

// warning! pixel [0, 15] is always transparent
procedure TransBlt(Canvas: TCanvas; X, Y: Integer; hBmp: HBitmap);
var
  BitmapDC: HDC;
  X1, Y1, Cur, Trans: Integer;
begin
  BitmapDC:= CreateCompatibleDC(Canvas.Handle);
  SelectObject(BitmapDC, hBmp);
  Trans:= GetPixel(BitmapDC, 0, 15);
  for X1:= 0 to 15 do
    for Y1:= 0 to 15 do begin
      Cur:= GetPixel(BitmapDC, X1, Y1);
      if (Cur <> Trans) then
        SetPixel(Canvas.Handle, X + X1, Y + Y1, Cur);
    end;

  DeleteDC(BitmapDC);
end;

procedure DisabledBlt(Canvas: TCanvas; X, Y: Integer; hBmp: HBitmap);
var
  XOffset: Integer;
  BitmapDC: HDC;
  Trans: TColor;

  procedure PaintColor(Color: TColor; Offset: Integer);
  var
    X1, Y1, Cur: Integer;
  begin
    for X1:= XOffset to XOffset + 15 do
      for Y1:= 0 to 15 do begin
        Cur:= GetPixel(BitmapDC, X1, Y1);

        if (Cur <> Trans)
        and (Cur and $000000FF <= 132)
        and (Cur and $0000FF00 shr 08 <= 128)
        and (Cur and $00FF0000 shr 16 <= 132)
        then
          SetPixel(Canvas.Handle, X + X1 + Offset, Y + Y1 + Offset, Color);
      end;
  end;

begin
  XOffset:= 0;
  BitmapDC:= CreateCompatibleDC(Canvas.Handle);
  SelectObject(BitmapDC, hBmp);
  Trans:= GetPixel(BitmapDC, 0, 15);
  PaintColor(ColorToRgb(clBtnHighlight), 1);
  PaintColor(ColorToRgb(clBtnShadow), 0);
  DeleteDC(BitmapDC);
end;

procedure CheckShowHint(MenuItem: Menus.TMenuItem; ShowFloatingHint: Boolean);
var
  S: String;
begin
  if not Assigned(MenuItem) then Exit;

//  if (MenuItem.Hint <> '') then
//    Application.Hint:= MenuItem.Hint;

  // thanks to Victor Santos
  if Assigned(ActivePopupMenu98)
  and not Assigned(ActivePopupMenu98.StatusBar)
  then Application.Hint:= MenuItem.Hint;

  if Assigned(ActivePopupMenu98)
  and (MenuItem.Enabled)
  then begin
    if Assigned(ActivePopupMenu98.StatusBar)
    then SetStatusBarText(MenuItem.Hint);

    if ShowFloatingHint
    and ActivePopupMenu98.ShowHint then begin
      S:= Trim(GetShortHint(MenuItem.Hint));

      Application.CancelHint;
      ActiveHintWindow98.ReleaseHandle;

      ActiveHintWindow98.ShowHint(S);
    end;
  end;
end;

function CreateSystemMenu98(AForm, AMC: TComponent; SystemType: TSystemType;
  Menu: TPopupMenu): TPopupMenu;

const
  MenuCmds: array [0..5] of Integer = (
    sc_Restore, sc_Move, sc_Size, sc_Minimize,
    sc_Maximize, sc_Close);

  MenuNames: array [0..5] of String = (
    'SYSTEMRESTORE', 'SYSTEMMOVE', 'SYSTEMSIZE', 'SYSTEMMINIMIZE',
    'SYSTEMMAXIMIZE', 'SYSTEMCLOSE');

var
  SysMenu: THandle;

  procedure AppendMenuItems(Items: Menus.TMenuItem; Menu: THandle);
  var
    I, P: Integer;
    M: TMenuItem;
    DefPresent: Boolean;
  begin
    DefPresent:= False;
    for I:= 0 to GetMenuItemCount(Menu) do
    with AMC as Forms.TForm do begin
      mii.dwTypeData:= Z;
      mii.cch:= SizeOf(Z) -1;
      if not GetMenuItemInfo(SysMenu, I, True, mii) then Continue;


      // create menu item for Delphi
      M:= TMenuItem.Create(AForm);


      // defining the current state
      if ((WindowState = wsNormal) and (mii.wID = sc_Restore))
      or ((WindowState = wsMaximized) and ((mii.wID = sc_Maximize) or (mii.wID = sc_Move) or (mii.wID = sc_Size)))
      then M.Enabled:= False;

      if ((mii.wID = sc_Close) and (not DefPresent)) or (mii.fState and mfs_Default <> 0)
      then M.Default:= True;


      // identification
      M.Tag:= mii.wID;


      // action
      M.OnClick:= TForm(AMC).SystemClick;


      // default already present?
      DefPresent:= DefPresent or (mii.fState and mfs_Default <> 0);

      if mii.fType and mft_Separator <> 0
      then M.Caption:= '-'
      else begin
        P:= Pos(#9, Z);
        if P <> 0 then begin
          M.Caption:= Copy(Z, 1, P -1);
          M.ShortCut:= TextToShortCut(Copy(Z, P +1, MaxInt));
        end
        else
          M.Caption:= Z;

        for P:= 0 to 5 do
          if mii.wID = MenuCmds[P] then begin
            AForm.FindComponent(MenuNames[P]).Free;
            M.Name:= MenuNames[P];
            Break;
          end;
      end;

      // submenus
      if mii.hSubMenu <> 0 then AppendMenuItems(M, mii.hSubMenu);
      Items.Add(M);
    end;
  end;

begin
  Result:= nil;
  if (not Assigned(AForm))
  or (not Assigned(AMC))
  then Exit;

  SysMenu:= GetSystemMenu(TWinControl(AMC).Handle, False);
  if (SysMenu = 0) then Exit;

  if Assigned(Menu) then begin
    Result:= Menu;
    while Result.Items.Count > 0 do Result.Items[0].Free;

  end
  else
    Result:= TPopupMenu98.Create(AForm);

//  TPopupMenu98(Result).MenuHandle:= Result.Items.Handle;

  // adding menu items
  mii.fMask:= miim_Type + miim_State + miim_ID + miim_SubMenu + miim_CheckMarks;
  AppendMenuItems(Result.Items, SysMenu);
end;

procedure UpdateMenuItemBitmaps(Form: Forms.TForm);
var
  I: Integer;
begin
  if not Assigned(Form) then Exit;

  for I:= 0 to Form.ComponentCount -1 do
    if Form.Components[I] is TMenuItemBitmaps then
      TMenuItemBitmaps(Form.Components[I]).Update;
end;

function LoadBitmapByName(S: String): HBitmap;
var
  J: Integer;
begin
  J:= 1;
  while J <= Length(S) do
    if (S[J] >= 'A') and (S[J] <='Z') then Inc(J)        else
    if (S[J] >= 'a') and (S[J] <='z') then Dec(S[J], 32) else
      System.Delete(S, J, 1);

  Result:= LoadBitmap(hInstance, PChar('AM98_' + S));
end;


// thanks to Jordan Russel again...
procedure CreatePattern;
var
  X, Y: Integer;
begin
  Pattern:= TBitmap.Create;
  Pattern.Width:= 8;
  Pattern.Height:= 8;
  for Y:= 0 to 7 do
    for X:= 0 to 7 do
      if (Y mod 2) = (X mod 2)
      then Pattern.Canvas.Pixels[X, Y]:= clBtnHighlight
      else Pattern.Canvas.Pixels[X, Y]:= clBtnFace;
end;

function RemoveAmpersands(S: String): String;
begin
  Result:= S;
  while Pos('&', Result) <> 0 do
    Delete(Result, Pos('&', Result), 1);
end;

// thanks to Brad Stowers for this routine
procedure PaintMenuIcon(Owner, AMC: Forms.TForm; DC: HDC);
var
  IconHandle, NewIcon: HIcon;
  Size: Integer;
begin
  if Assigned(AMC) and (AMC.Icon.Handle <> 0) then
     IconHandle := AMC.Icon.Handle
  else
  if Assigned(AMC) and (Owner.Icon.Handle <> 0) then
    IconHandle:= Owner.Icon.Handle
  else
  if Application.Icon.Handle <> 0 then
    IconHandle:= Application.Icon.Handle
  else
    IconHandle:= LoadIcon(0, IDI_APPLICATION);

  Size:= GetSystemMetrics(SM_CXSMICON);
  NewIcon:= CopyImage(IconHandle, IMAGE_ICON, Size, Size, $4000);
  DrawIconEx(DC, 0, 0, NewIcon, 0, 0, 0, 0, DI_NORMAL);

  DeleteObject(NewIcon);
end;


procedure Register;
begin
  RegisterComponents('Toolbar97', [TMainMenu98, TPopupMenu98, TMenuItemBitmaps, TSystemMenu98, TOnlyMainMenu]);
end;


{ overriden routines }
{ this prevents memory leaks }

function NewSubMenu(const ACaption: string; hCtx: Word; const AName: string;
  Items: array of Menus.TMenuItem): TMenuItem;
var
  I: Integer;
begin
  Result:= TMenuItem.Create(nil);
  for I:= Low(Items) to High(Items) do
    Result.Add(Items[I]);
  Result.Caption:= ACaption;
  Result.HelpContext:= hCtx;
  Result.Name:= AName;
end;

function NewItem(const ACaption: string; AShortCut: TShortCut;
  AChecked, AEnabled: Boolean; AOnClick: TNotifyEvent; hCtx: Word;
  const AName: string): TMenuItem;
begin
  Result:= TMenuItem.Create(nil);
  with Result do begin
    Caption:= ACaption;
    ShortCut:= AShortCut;
    OnClick:= AOnClick;
    HelpContext:= hCtx;
    Checked:= AChecked;
    Enabled:= AEnabled;
    Name:= AName;
  end;
end;

function NewLine: TMenuItem;
begin
  Result:= TMenuItem.Create(nil);
  Result.Caption:= '-';
end;


{ TMainMenu98 }

constructor TMainMenu98.Create(AOwner: TComponent);
var
  I: Integer;
begin
  inherited Create(AOwner);

  Caption:=       'Menu Bar';
  FullSize:=      True;
  FSubMenus:=     TSubmenus.Create(Self);
  FOptions:=      TOptions.Create(nil);
  FGrayInactive:= True;
  FOptions.FTimeout:= 10;
  FMergedMenusList:=  TList.Create;

  if (csDesigning in ComponentState) then
    for I:= 0 to Owner.ComponentCount -1 do begin
      if Owner.Components[I] is TMainMenu then
        FMainMenu:= Owner.Components[I] as TMainMenu;
    end
  else
    if (Forms.TForm(Owner).FormStyle = fsMdiForm) then begin
      Timer:= TTimer.Create(Owner);
      Timer.Interval:= 100;
      Timer.OnTimer:= CheckForMDIChildMenu;
    end;

  ActiveMenu98List.Add(Self);
end;


destructor TMainMenu98.Destroy;
begin
  Timer.Free;
  FSubMenus.Free;
  FOptions.Free;
  ActiveMenu98List.Remove(Self);
  FMergedMenusList.Free;

  inherited;
end;

procedure TMainMenu98.Loaded;
var
  I: Integer;
begin
  inherited;

  if (csDesigning in ComponentState)
  or not Assigned(FMainMenu) then Exit;

  while ControlCount > 0 do
    Controls[ControlCount -1].Free;

  for I:= 0 to FMainMenu.Items.Count -1 do
    if FMainMenu.Items[I].Visible then
      InsertControl(CreateButtonForMenu(FMainMenu.Items[I], Owner, stAutomatic));

  HideWindowMenu(Owner);
end;

procedure TMainMenu98.SetMenuProperties(Menu: TPopupMenu);
begin
  if not (Menu is TPopupMenu98) then Exit;

  with TPopupMenu98(Menu) do begin
    Options.Assign(Self.Options);
    Submenus.Assign(Self.Submenus);
    Font.Assign(Self.Font);

    StatusBar:=       Self.StatusBar;
    StatusBarIndex:=  Self.FSBPanelNo;
    ParentFont:=      Self.ParentFont;
    SystemFont:=      Self.SystemFont;
    ShowHint:=        Self.ShowHint;
    ParentShowHint:=  Self.ParentShowHint;

    OnPopup:=         Self.OnPopup;
    OnMenuCommand:=   Self.OnMenuCommand;
    OnMenuClose:=     Self.OnMenuClose;
  end;
end;

function TMainMenu98.CreateButtonForMenu(MenuItem: Menus.TMenuItem; AOwner: TComponent;
  SystemType: TSystemType): TToolbarButton98;
var
  B: TToolbarButton98;
begin
  B:= TToolbarButton98.Create(Self);
  B.Font:= Font;
  B.ParentFont:= False;
  B.DropdownArrow:= False;
  B.AllowAllUp:= True;
  B.ShowHint:= ShowHint;
  B.ParentShowHint:= False;
  B.Cursor:= Cursor;
  B.SystemButton:= SystemType <> stAutomatic;
  B.AutoSize:= True;

  if Assigned(MenuItem) then begin
    B.Caption:= MenuItem.Caption;
    B.Enabled:= MenuItem.Enabled;
    B.Hint:=    MenuItem.Hint;
    B.OnClick:= MenuItem.OnClick;

    if MenuItem.Default then
      B.Font.Style:= B.Font.Style + [fsBold];
  end;

  mii.fMask:= miim_CheckMarks;
  if Assigned(MenuItem)
  and Assigned(MenuItem.Parent)
  and GetMenuItemInfo(MenuItem.Parent.Handle, MenuItem.Command, False, mii)
  and (mii.hbmpUnchecked <> 0)
  then B.Glyph.Handle:= mii.hbmpUnchecked;

  if Assigned(MenuItem)
  and (MenuItem.Count > 0)
  then begin
    B.GroupIndex:= 1;
    B.DropdownMenu:= TPopupMenu98.Create(AOwner);
    SetmenuProperties(B.DropdownMenu);
    TPopupMenu98(B.DropdownMenu).MenuItems:= MenuItem;
  end;

  Result:= B;
end;

procedure TMainMenu98.MergeMenus;

  function CheckForGroupIndex(Index: Integer; MenuItems: Menus.TMenuItem): Boolean;
  var
    I: Integer;
  begin
    Result:= False;
    for I:= 0 to MenuItems.Count -1 do
      if MenuItems[I].Visible
      and (MenuItems[I].GroupIndex = Index) then begin
        InsertControl(CreateButtonForMenu(MenuItems[I], Owner, stAutomatic));
        Result:= True;
      end;
  end;

var
  AMC: Forms.TForm;
  F, G: Boolean;
  I: Integer;
begin
  G:= False;
  FMergedMenusList.Clear;
  AMC:= Forms.TForm(Owner).ActiveMdiChild;
  F:= Assigned(AMC) and Assigned(AMC.Menu);

  // update menu item bitmaps
  UpdateMenuItemBitmaps(Forms.TForm(Owner));
  UpdateMenuItemBitmaps(AMC);

  // remove all controls
  while ControlCount > 0 do begin
    if Controls[0] is TToolbarButton98 then
      with Controls[0] as TToolbarButton98 do begin
        Glyph.ReleaseHandle;

        if Assigned(DropdownMenu) then
          TPopupMenu98(DropdownMenu).Form.Release;
      end;

    Controls[0].Free;
  end;

  // reinsert menu buttons
  for I:= 0 to 255 do begin
    if F then
      G:= CheckForGroupIndex(I, AMC.Menu.Items);

    if (not F) or (not G) then
      CheckForGroupIndex(I, FMainMenu.Items);
  end;

  if F then
    FMergedMenusList.Add(AMC.Menu);

end;

procedure TMainMenu98.Rebuild;
begin
  if Updating then Exit;
  BeginUpdate;
  EndUpdate;
end;

procedure TMainMenu98.BeginUpdate;
begin
  if Updating then Exit;
  Updating:= True;

  inherited BeginUpdate;

//  LockDesktop;
end;

procedure TMainMenu98.EndUpdate;
begin
  if not Updating then Exit;

  KillActivePopupMenu98(False);

  MergeMenus;
  EndUpdate1;
end;

procedure TMainMenu98.EndUpdate1;
// sometimes I don't need to merge menus again and just call this function directly
var
  I: Integer;
  AMC: Forms.TForm;
  B: TToolbarButton98;
  S: TSpeedButton;
  P: TPoint;
  L: TList;
begin
  if not Updating then Exit;

  AMC:= Forms.TForm(Owner).ActiveMdiChild;

  // remove all
  I:= 0;
  SpaceButton:= nil;
  while I < ControlCount do begin
    if (Controls[I] is TToolbarButton98)
    and TToolbarButton98(Controls[I]).SystemButton then begin
      TToolbarButton98(Controls[I]).Glyph.ReleaseHandle;
      Controls[I].Free;
      Continue;
    end;

    if Controls[I] is TSpeedButton then begin
      Controls[I].Free;
      Continue;
    end;

    Inc(I);
  end;


  if Assigned(AMC)
  and (AMC.WindowState = wsMaximized) then begin

    // create system menu button
    if (AMC is Forms.TForm) then begin

      // create button
      B:= CreateButtonForMenu(nil, Owner, stMdiChildSystem);
      B.DropdownMenu:= CreateSystemMenu98(Owner, AMC, stMdiChildSystem, nil);
      SetMenuProperties(B.DropdownMenu);
      B.DisplayMode:= dmGlyphOnly;
      B.Width:= 20;
      B.Tag:= sc_Close;
      B.OnDblClick:= MenuSystemClick;

      B.Glyph.Width:= 17;
      B.Glyph.Height:= 17;

      with B.Glyph.Canvas do begin
        Brush.Color:= Self.Color;
        FillRect(ClipRect);
        PaintMenuIcon(Forms.TForm(Owner), AMC, Handle);
      end;



      // now we have set B to the first control in Controls array.
      // This removes all controls and insertis it again.
      // Really stupid, I think
      L:= TList.Create;
      while ControlCount > 0 do begin
        L.Add(Controls[0]);
        RemoveControl(Controls[0]);
      end;

      B.Parent:= Self;
      while L.Count > 0 do begin
        InsertControl(TControl(L[0]));
        L.Delete(0);
      end;

      L.Free;
    end;

    // space between buttons
    SpaceButton:= TSpeedButton.Create(Self);
    SpaceButton.Enabled:= False;
    SpaceButton.Parent:= Self;

    if (biMaximize in AMC.BorderIcons)
    or (biMinimize in AMC.BorderIcons) then begin
      // minimize button
      S:= TSpeedButton.Create(Self);
      S.Width:= 16;
      S.Height:= 14;
      S.Glyph.Handle:= LoadBitmap(HInstance, 'MINIMIZE_SMALL');
      S.Enabled:= biMinimize in AMC.BorderIcons;
      S.Tag:= sc_Minimize;
      S.OnClick:= MenuSystemClick;
      S.Hint:= 'Minimize window';
      S.ShowHint:= ShowHint;
      S.Parent:= Self;

      // restore button
      S:= TSpeedButton.Create(Self);
      S.Width:= 16;
      S.Height:= 14;
      S.Glyph.Handle:= LoadBitmap(HInstance, 'RESTORE_SMALL');
      S.Enabled:= biMaximize in AMC.BorderIcons;
      S.Tag:= sc_Restore;
      S.OnClick:= MenuSystemClick;
      S.Hint:= 'Restore window';
      S.ShowHint:= ShowHint;
      S.Parent:= Self;

      // space between buttons
      S:= TSpeedButton.Create(Self);
      S.Height:= 0;
      S.Width:= 2;
      S.Enabled:= False;
      S.Parent:= Self;
    end;

    // close button
    S:= TSpeedButton.Create(Self);
    S.Width:= 16;
    S.Height:= 14;
    S.Glyph.Handle:= LoadBitmap(HInstance, 'CLOSE_SMALL');
    S.Tag:= sc_Close;
    S.OnClick:= MenuSystemClick;
    S.Hint:= 'Close window';
    S.ShowHint:= ShowHint;
    S.Parent:= Self;

    P:= Point(0, 0);
    OrderControls(True, nil, nil, P);
    SetBounds(Left, Top, 0, 0);
  end;

  LastButtonMouseInControl:= nil;

  inherited EndUpdate;

  if Application.Active
  and Forms.TForm(Owner).Visible
  then Forms.TForm(Owner).SetFocus;

  Updating:= False;
  HideWindowMenu(Owner);
//  UnlockDesktop;
  ProcessPaintMessages;
end;

procedure TMainMenu98.EasyUpdate(MenuItem: Menus.TMenuItem);
// calls when TMenuItem.Enabled, Caption
// and other properties is changed
var
  I: Integer;
begin
  // update all properties - looks terrific, right?
  for I:= 0 to ControlCount -1 do
    if Controls[I] is TToolbarButton98 then
      with TToolbarButton98(Controls[I]) do
        if Assigned(DropdownMenu) then
          with TPopupMenu98(DropdownMenu) do
            if MenuItems = MenuItem then begin
              Caption:= MenuItem.Caption;
              Enabled:= MenuItem.Enabled;
              Hint:=    MenuItem.Hint;
              OnClick:= MenuItem.OnClick;

              if MenuItem.Default
              then Font.Style:= Font.Style + [fsBold]
              else Font.Style:= Font.Style - [fsBold];

              HBState:= False;
              VBState:= False;
              SetBounds(Left, Top, 0, 0);
              Paint;
              Break;
            end;

  inherited EndUpdate;

  Updating:= False;
  HideWindowMenu(Owner);
//  UnlockDesktop;
end;

procedure TMainMenu98.MenuSystemClick(Sender: TObject);
var
  AMC: Forms.TForm;
begin
  AMC:= Forms.TForm(Owner).ActiveMdiChild;
  SendMessage(AMC.Handle, wm_SysCommand, TComponent(Sender).Tag, 0);
end;

procedure TMainMenu98.SetBounds(ALeft, ATop, AWidth, AHeight: Integer);
var
  P: TPoint;
  I, Sum: Integer;
begin
  if (AWidth <> 0)
  and (AHeight <> 0)
  and (not Assigned(DockedTo))
  and (ControlCount > 0) then begin
    P:= Point(0, 0);
    OrderControls(True, nil, nil, P);
  end;

  // arrange panel
  if Assigned(SpaceButton) then
    if Assigned(DockedTo)
    and (DockedTo.Position in [dpTop, dpBottom]) then begin
      Sum:= 0;
      for I:= 0 to ControlCount -1 do
        if Controls[I] <> SpaceButton then
          Inc(Sum, Controls[I].Width);

      SpaceButton.Height:= 0;
      SpaceButton.Width:= DockedTo.Width - Sum -13;
    end
    else begin
      SpaceButton.Height:= 1;
      SpaceButton.Width:= 1;
    end;

  inherited;
end;

procedure TMainMenu98.SetMainMenu(Value: TMainMenu);
begin
  if not Assigned(Value) then
    raise Exception.Create('Value cannot be empty');

  FMainMenu:= Value;

  if (csDesigning in ComponentState)
  or (csLoading in ComponentState)
  then Exit;

  Rebuild;
end;

procedure TMainMenu98.CheckForMDIChildMenu(Sender: TObject);
const
  LastAMC:    Forms.TForm  = nil;
  LastAMCWS:  TWindowState = wsNormal;
  LastAMCMnu: TMainMenu    = nil;

var
  AMC: Forms.TForm;
  WindowStateChanged, MenuChanged, MaximizedChildClosed: Boolean;
begin
  AMC:= Forms.TForm(Owner).ActiveMdiChild;

  WindowStateChanged:= Assigned(AMC)
    and (LastAMCWS <> AMC.WindowState)
    and ((AMC.WindowState = wsMaximized) or (LastAMCWS = wsMaximized));

  MenuChanged:= Assigned(AMC) and (LastAMCMnu <> AMC.Menu);

  MaximizedChildClosed:= (not Assigned(AMC))
    and (LastAMCWS = wsMaximized);


  if WindowStateChanged or MenuChanged or MaximizedChildClosed then begin
    BeginUpdate;

    if MenuChanged
    then MergeMenus;

    EndUpdate1;
    
  end;

  LastAMC:= AMC;
  if Assigned(AMC) then begin
    LastAMCWS:= AMC.WindowState;
    LastAMCMnu:= AMC.Menu;
  end
  else begin
    LastAMCWS:= wsNormal;
    LastAMCMnu:= nil;
  end;
end;

procedure TMainMenu98.SetSubMenus(Value: TSubMenus);
begin
  FSubMenus.Assign(Value);
end;


procedure TMainMenu98.SetSystemFont(Value: Boolean);
begin
  FSystemFont:= Value;
  if Value then ParentFont:= False;
end;

function TMainMenu98.IsFontStored: Boolean;
begin
  Result:= not (ParentFont or SystemFont);
end;



{ TToolbarButton98 }

constructor TToolbarButton98.Create(AOwner: TComponent);
begin
  inherited;
  HBuffer:= TBitmap.Create;
  VBuffer:= TBitmap.Create;
end;

destructor TToolbarButton98.Destroy;
begin
  HBuffer.Free;
  VBuffer.Free;
  inherited;
end;

procedure TToolbarButton98.CMDialogChar(var Msg: TCMDialogChar);
begin
  if ((Msg.KeyData and $20000000) = 0) then
    if not Assigned(ButtonMouseInControl)
    then Exit
    else
      if IsAccel(Msg.CharCode, Caption) and Enabled and Visible then
        ButtonMouseInControl.MouseLeft;

  inherited;
end;

procedure TToolbarButton98.MouseMove(Shift: TShiftState; X, Y: Integer);
const
  LastX: Integer = 0;
  LastY: Integer = 0;
  PleaseClickMe: Boolean = False;
begin
  if PleaseClickMe then begin
    PleaseClickMe:= False;

    // remove SelectedIndex
    if Assigned(DropdownMenu)
    and (DropdownMenu is TPopupMenu98) then
      with TPopupMenu98Form(TPopupMenu98(DropdownMenu).Form) do
        if SelectedIndex <> -1 then FSelectedIndex:= -1;

    if Enabled then Click;
    Exit;
  end;

  // ignore some little movements
  if (Abs(X - LastX) <= 2)
  and (Abs(Y - LastY) <= 2)
  then Exit;

  LastX:= X;
  LastY:= Y;

  if (not Down)
  and Assigned(DropdownMenu)
  and AssignedActivePopupMenu98Form
  and (ActivePopupMenu98.Button <> nil)
  and (ActivePopupMenu98.Button <> Self)
  then begin
    KillActivePopupMenu98(True);
    PleaseClickMe:= True;
  end;

  inherited MouseMove(Shift, X, Y);

  if AssignedActiveMenu98
  and Assigned(ActiveMenu98.StatusBar)
  then SetStatusBarText(Hint);

  LastButtonMouseInControl:= Self;
end;

procedure TToolbarButton98.MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  // on mouse click remove SelectedIndex
  if Assigned(DropdownMenu)
  and (DropdownMenu is TPopupMenu98) then
    with TPopupMenu98Form(TPopupMenu98(DropdownMenu).Form) do
      if SelectedIndex <> -1 then FSelectedIndex:= -1;

  inherited;
end;

procedure TToolbarButton98.Click;
begin
  if AssignedActiveMenu98 then
    ActiveMenu98.KeepSelected:= False;

  if (Down or SystemButton)
  and AssignedActivePopupMenu98Form
  and (ActivePopupMenu98 = DropdownMenu) then begin
    KillActivePopupMenu98(True);
    Down:= False;
    MouseEntered;
    ShowCaret(0);

  end
  else begin
    KillActivePopupMenu98(True);
    ButtonMouseInControl:= nil;
    LastButtonMouseInControl:= Self;

    // event
    if Assigned(OnClick)
    then OnClick(Self);

    if (Assigned(DropdownMenu)
    and (DropdownMenu is TPopupMenu))
    then begin
      Down:= True;
      Paint;

      if (DropdownMenu is TPopupMenu98) then begin
        MouseLeft;
        MouseCapture:= False;
        if Assigned(DropdownMenu) then
        with TPopupMenu98(DropdownMenu) do begin
          PopupComponent:= Self;
          TMainMenu98(Parent).SetMenuProperties(DropdownMenu);
        end;

        // alignment
        PP:= Point(0, Height);
        if (Parent is TCustomToolWindow97)
        and AssignedActiveMenu98 then
          case GetDockTypeOf(TMainMenu98(Parent).DockedTo) of
            dtLeftRight:
              PP:= Point(Width, 0);
            dtNotDocked:
              if (Top + Height + 25 < ActiveMenu98.Height)
              and ((Left + Width * 2) > ActiveMenu98.Width)
              then PP:= Point(Width, 0);
          end;

        PP:= ClientToScreen(PP);
        if SystemButton then begin
          Timer.Free;
          Timer:= TTimer.Create(Self);
          Timer.Interval:= 175;
          Timer.OnTimer:= DelaySystemMenu;
        end
        else begin
          DropdownMenu.Popup(PP.X, PP.Y);
          Down:= TPopupMenu98(DropdownMenu).FormOnScreen;
        end;
      end
      else
        ShowCaret(0);

    end
    else begin
      ShowCaret(0);
      MouseLeft;
      ButtonMouseInControl:= nil;
      sndPlaySound(PChar(MenuCommandSound), snd_Async + snd_NoDefault);
      MouseEntered;

    end;
  end;
end;

procedure TToolbarButton98.DelaySystemMenu(Sender: TObject);
var
  F: TComponent;
  AMC: Forms.TForm;
begin
  Timer.Free;
  Timer:= nil;

  // create menu
  F:= Owner;
  while not (F is Forms.TForm) do
    F:= F.Owner;

  AMC:= Forms.TForm(F).ActiveMdiChild;
  DropdownMenu:= CreateSystemMenu98(F, AMC, stMdiChildSystem, DropdownMenu);

  if Owner is TMainMenu98 then
    TMainMenu98(Owner).SetMenuProperties(DropdownMenu);

  // display it
  if Assigned(DropdownMenu) then
    DropdownMenu.Popup(PP.X, PP.Y);

end;

procedure TToolbarButton98.WMLButtonDblClk(var Msg: TWMMouse);
begin
  if SystemButton
  then begin
    if Assigned(Timer)
    then Timer.Enabled:= False;
    
    Timer.Free;
    KillActivePopupMenu98(False);
    TForm(ActiveMenu98.Owner).ActiveMdiChild.Close;
  end
  else inherited;
end;

procedure TToolbarButton98.cmMouseLeave(var Msg: TMessage);
begin
  if not (AssignedActiveMenu98
  and ActiveMenu98.KeepSelected) then
    inherited;
end;

procedure TToolbarButton98.cmMouseEnter(var Msg: TMessage);
begin
  inherited;
  if not Enabled then MouseMove([], -1, -1);
end;


procedure TToolbarButton98.SetBounds(ALeft, ATop, AWidth, AHeight: Integer);
var
  I, H, W: Integer;
  S: String;
begin
  if FAutoSize then try
    if (Parent is TToolbar97)
    and Assigned(Canvas)
    and Assigned(Glyph) then begin
      I:= 1;
      S:= Caption;
      while I <= Length(S) do begin
        if S[I] = '&' then begin
          Delete(S, I, 1);
          if (I <= Length(S)) and (S[I] = '&') then Inc(I); // if '&&' found...
        end;
        Inc(I);
      end;

      I:= SelectObject(Canvas.Handle, Font.Handle);

      if SystemButton
      then begin H:= 19; W:= 20; end
      else begin
        H:= Canvas.TextHeight(S) +6;
        W:= Canvas.TextWidth(S) + Glyph.Height + 16;
      end;

      SelectObject(Canvas.Handle, I);

      if Assigned(TToolbar97(Parent).DockedTo)
      and (TToolbar97(Parent).DockedTo.Position in [dpLeft, dpRight])
      then begin
        AWidth:= H; AHeight:= W;
      end
      else begin
        AWidth:= W; AHeight:= H;
        if Assigned(TToolbar97(Parent).DockedTo) then ATop:= 0;
      end;

    end;
  except
  end;

  inherited;
end;

procedure TToolbarButton98.Paint;
const
  dt_Flags = dt_Center + dt_VCenter + dt_SingleLine + dt_NoClip;

var
  R: TRect;
  GlyphPos: TPoint;
  X, Y, old: Integer;

  procedure CreateRotatedBitmap;
  var
    W, C, P, X, Y: Integer;
  begin
    VBuffer.Width:=  HBuffer.Height;
    VBuffer.Height:= HBuffer.Width;
    C:= ColorToRgb(Color);
    W:= HBuffer.Height;

    with VBuffer.Canvas do begin
      Brush.Color:= Color;
      Brush.Style:= bsSolid;
      FillRect(ClipRect);
    end;

    for X:= 0 to HBuffer.Width -1 do
      for Y:= 0 to HBuffer.Height -1 do begin
        P:= HBuffer.Canvas.Pixels[X, Y];
        if P <> C then // not a background color
          VBuffer.Canvas.Pixels[W - Y -1, X]:= P;
      end;
  end;


begin
  if (FState = bsMouseIn)
  and (Assigned(ActivePopupMenu98))
  then MouseLeft;

  if (FState <> bsUp)
  and SystemButton
  then FState:= bsUp;

  if (FState = bsUp)
  and AssignedActiveMenu98
  and TMainMenu98(Owner).KeepSelected
  and (Application.Active)
  and (LastButtonMouseInControl = Self)
  and (ButtonMouseInControl = nil)
  then begin
    ButtonMouseInControl:= Self;
    MouseEntered;
  end;

  if (csDesigning in ComponentState) then begin
    HBState:= False;
    VBState:= False;
  end;
  

  if not HBState then begin
    // bounds
    if not IsVertical then begin
      HBuffer.Width:= Width;  HBuffer.Height:= Height;
    end
    else begin
      HBuffer.Width:= Height;  HBuffer.Height:= Width;
    end;

    // clearing the client area
    R:= HBuffer.Canvas.ClipRect;
    HBuffer.Canvas.Brush.Color:= Color;
    HBuffer.Canvas.Brush.Style:= bsSolid;
    HBuffer.Canvas.FillRect(R);

    if SystemButton
    then GlyphPos:= Point(2, 0)
    else GlyphPos:= Point(6, 0);

    if IsVertical
    then GlyphPos.Y:= Width  div 2 -7
    else GlyphPos.Y:= Height div 2 -7;

    // draw the button
    Inc(R.Left, Glyph.Width);
    HBuffer.Canvas.Brush.Style:= bsClear;
    old:= SelectObject(HBuffer.Canvas.Handle, Font.Handle);

    if Enabled then begin
      if (Glyph.Width <> 0) and (Glyph.Height <> 0) then
        TransBlt(HBuffer.Canvas, GlyphPos.X, GlyphPos.Y, Glyph.Handle);

      DrawText(HBuffer.Canvas.Handle, PChar(Caption), Length(Caption), R, dt_Flags);
    end
    else begin
      if (Glyph.Width <> 0) and (Glyph.Height <> 0) then
        DisabledBlt(HBuffer.Canvas, GlyphPos.X, GlyphPos.Y, Glyph.Handle);

      OffsetRect(R, 1, 1);
      SetTextColor(HBuffer.Canvas.Handle, ColorToRgb(clBtnHighlight));
      DrawText(HBuffer.Canvas.Handle, PChar(Caption), Length(Caption), R, dt_Flags);

      OffsetRect(R, -1, -1);
      SetTextColor(HBuffer.Canvas.Handle, ColorToRgb(clBtnShadow));
      DrawText(HBuffer.Canvas.Handle, PChar(Caption), Length(Caption), R, dt_Flags);
    end;

    SelectObject(HBuffer.Canvas.Handle, old);

    HBState:= True;
  end;

  // generate vertical bitmap
  if IsVertical and not VBState then begin
    CreateRotatedBitmap;
    VBState:= True;
  end;


  // drawing
  X:= 0;  Y:= 0;
  if (FState = bsDown) or (FState = bsExclusive)
  then begin Inc(X); Inc(Y); end;

  if IsVertical
  then Canvas.Draw(X, Y, VBuffer)
  else Canvas.Draw(X, Y, HBuffer);

  // border
  R:= Canvas.ClipRect;

  if (FState = bsMouseIn) or (csDesigning in ComponentState)
  then DrawEdge(Canvas.Handle, R, bdr_RaisedInner, bf_Rect);

  if (FState = bsDown) or (FState = bsExclusive)
  then DrawEdge(Canvas.Handle, R, bdr_SunkenOuter, bf_Rect);
end;

function TToolbarButton98.IsVertical: Boolean;
begin
  Result:= (FOrientation = orVertical)
    or ((FOrientation = orAutomatic) and (Height > Width));
end;

procedure TToolbarButton98.SetOrientation(Value: TButtonOrientation);
begin
  FOrientation:= Value;
  Refresh;
end;

procedure TToolbarButton98.SetAutoSize(Value: Boolean);
begin
  FAutoSize:= Value;
  if Value then SetBounds(Left, Top, Width, Height);
end;

procedure TToolbarButton98.CMSysColorChange(var Msg: TMessage);
begin
  // buffers' state
  HBState:= False;
  VBState:= False;

  // pattern for checkmarks
  Pattern.Free;
  Pattern:= nil;

  inherited;
end;


{ TPopupMenu98 }

constructor TPopupMenu98.Create(AOwner: TComponent);
begin
  inherited;

  FOptions:= TOptions.Create(nil);
  FSubmenus:= TSubmenus.Create(Self);
  FFont:= TFont.Create;
  FParentFont:= True;

  // this form will be automatically destroyed by the Owner
  if not (csDesigning in ComponentState) then
    Form:= TPopupMenu98Form.Create(Owner);
end;

destructor TPopupMenu98.Destroy;
begin
  FOptions.Free;
  FSubmenus.Free;
  FFont.Free;

  inherited;
end;

procedure TPopupMenu98.Popup(X, Y: Integer);
var
  H: THandle;
begin
  if (PopupComponent is TToolbarButton97)
  and (not (PopupComponent is TToolbarButton98))
  and FormOnScreen
  then begin
    TPopupMenu98Form(Form).AnimatedHide;
    KillActivePopupMenu98(False);
    Exit;
  end;

  KillActivePopupMenu98(False);
  ActivePopupMenu98:= Self;

  // event
  if Assigned(OnPopup)
  then OnPopup(Self);

  Forms.TForm(Owner).BringToFront;

  with Form as TPopupMenu98Form do begin
    ParentPopupMenu98Form:= nil;

    // initialization
    if Assigned(Self.MenuItems)
    then MenuItems:= Self.MenuItems
    else MenuItems:= Self.Items;

    if Self.MenuHandle <> 0
    then MenuHandle:= Self.MenuHandle
    else MenuHandle:= MenuItems.Handle;

    if Button <> nil
    then Cursor:= Button.Cursor;

    if Self.FParentFont
    then
      Font:= Forms.TForm(Owner).Font
    else
    if Self.FSystemFont
    then begin
      h:= GetStockObject(Ansi_Var_Font);
      Font.handle:= h;
    end
    else
      Font:= Self.Font;

    OpenAnimation:=   Self.Options.OpenAnimation;
    CloseAnimation:=  Self.Options.CloseAnimation;
    PopupMenu:=       Self;
    SelectedChanged:= True;

    UpdateMenuItemBitmaps(Forms.TForm(Owner));

    // show menu
    if (not IsWindowEnabled(Handle))
    then EnableWindow(Handle, True);

    SetBounds(X, Y, 0, 0);
    BringMenuToFront;

    // animation
    try
      if Animate then begin

        // we should simpulate TrackPopupMenu for ToolbarButton97
        if (PopupComponent is TToolbarButton97)
        and not (PopupComponent is TToolbarButton98)
        then
          while FormOnScreen do
            Application.ProcessMessages;

        Exit;
      end;
    except
    end;

    SilentHide;
    if ActivePopupMenu98 = Self then ActivePopupMenu98:= nil;
  end;
end;

function TPopupMenu98.Button: TToolbarButton98;
begin
  if PopupComponent is TToolbarButton98
  then Result:= TToolbarButton98(PopupComponent)
  else Result:= nil;
end;

function TPopupMenu98.FormOnScreen: Boolean;
begin
  Result:= Assigned(Form)
//    and (Form.Visible)
    and (Form.Width <> 0)
    and (Form.Height <> 0);
end;

function TPopupMenu98.IsFontStored: Boolean;
begin
  Result:= not (FParentFont or FSystemFont);
end;

function TPopupMenu98.IsShowHintStored: Boolean;
begin
  Result:= not FParentShowHint;
end;

procedure TPopupMenu98.SetFont(Value: TFont);
begin
  FFont.Assign(Value);
end;

procedure TPopupMenu98.SetParentFont(Value: Boolean);
begin
  FParentFont:= Value;
  if Value then FSystemFont:= False;
end;

procedure TPopupMenu98.SetSystemFont(Value: Boolean);
begin
  FSystemFont:= Value;
  if Value then FParentFont:= False;
end;

procedure TPopupMenu98.SetShowHint(Value: Boolean);
begin
  if FShowHint <> Value then begin
    FShowHint:= Value;
    FParentShowHint:= False;
  end;
end;

procedure TPopupMenu98.SetParentShowHint(Value: Boolean);
begin
  if FParentShowHint <> Value then
    FParentShowHint:= Value;
end;

procedure TPopupMenu98.SetSubMenus(Value: TSubMenus);
begin
  FSubMenus.Assign(Value);
end;


{ TSystemMenu98 }

constructor TSystemMenu98.Create(AOwner: TComponent);
begin
  inherited;

  FEnabled:= True;
  FAllowAdjust:= True;

  if (csDesigning in ComponentState) then Exit;

  if (Owner is AnimatedMenus98.TForm)
  then
    TForm(Owner).SystemMenu98:= Self
  else
    raise Exception.Create('Error: Owner if TSystemMenu98 component is not an ' +
      'AnimatedMenus98.TForm. Possible reason: AnimatedMenus98 '+
      'unit doesn''t placed at the end of uses clause of your form''s unit');
end;



{ TPopupMenu98Form }

constructor TPopupMenu98Form.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);

  // own initializaton
  FSelectedIndex:= -1;
  LastSelectedIndex:= -1;
  OpenAnimation:=  anVSlide;
  CloseAnimation:= anVSlide;

  Timer:= TTimer.Create(Self);
  Timer.Enabled:= False;
  Timer.Interval:= 250;
  Timer.OnTimer:= PopupMenu98FormTimer;

  Buffer:= TBitmap.Create;
  Buffer.Canvas.Font.Assign(Font);

  SetBounds(0, 0, 0, 0);
  ShowWindow(Handle, sw_Show);

  // setting the hook - many thanks to Victor Santos
  // for help in solving problems with hook
  if HGetMessageHook = 0
  then HGetMessageHook:= SetWindowsHookEx(wh_GetMessage, @GetMessageHook, 0, GetCurrentThreadID);

end;

destructor TPopupMenu98Form.Destroy;
begin
  MenuItems:= nil;

  Timer.Free;
  Buffer.Free;

  ActiveHintWindow98.ReleaseHandle;

  // this checking is necessary when submenu going to be killed
  // by MainForm (Owner property), not by ParentPopupMenu98
  if Assigned(ParentPopupMenu98Form)
  and (ParentPopupMenu98Form.CurSubMenu = Self) then
    ParentPopupMenu98Form.CurSubMenu:= nil;

  if Assigned(CurSubMenu) then
    CurSubMenu.Release;

  inherited Destroy;
end;


{ Properties }

function TPopupMenu98Form.GetRealHeight: Integer;
var
  I: Integer;
  C: TComponent;
begin
  Result:= 0;
  if (MenuHandle = 0)
  or not Assigned(Options)
  then Exit;

  Result:= Options.ItemSpace;

  mii.fMask:= miim_Type;
  for I:= 0 to MenuItemCount -1 do begin
    mii.dwTypeData:= Z;
    mii.cch:= SizeOf(Z) -1;
    if not GetMenuItemInfo(MenuHandle, I, True, mii) then Continue;

    if mii.fType and mft_Separator <> 0
    then Inc(Result, Options.SeparatorHeight)
    else
     begin
      C:= nil;
      if (Z[0] = ':') and (Z[0] = ':') then
        C:= FindComponent(Copy(Z, 3, MaxInt));

      if Assigned(C)
      and (C is TControl)
      then Inc(Result, TControl(C).Height)
      else
        Inc(Result, Options.ItemHeight);
    end;

    Inc(Result, Options.ItemSpace);
  end;

  // fit to background
  I:= Options.Background.Height -3;
  if Options.FitToBackground
  and (Result < I)
  then Result:= I;

end;

function TPopupMenu98Form.GetRealWidth: Integer;
var
  S: String;
  I, P, W1, W2, R1, R2: Integer;
  C: TComponent;
begin
  R1:= 0;
  R2:= 0;
  Result:= 0;
  if MenuHandle = 0 then Exit;

  mii.fMask:= miim_Type + miim_State;
  for I:= 0 to MenuItemCount -1 do begin
    P:= 0;
    C:= nil;
    mii.dwTypeData:= Z;
    mii.cch:= SizeOf(Z) -1;
    if (not GetMenuItemInfo(MenuHandle, I, True, mii))
    or (mii.fType and mft_Separator <> 0)
    then Continue;

    if mii.fState and mfs_Default <> 0
    then Canvas.Font.Style:= [fsBold]
    else Canvas.Font.Style:= [];

    if (Z[0] = ':') and (Z[0] = ':') then
      C:= FindComponent(Copy(Z, 2, MaxInt));

    if Assigned(C)
    and (C is TControl)
    then
      W1:= TControl(C).Width
    else begin
      S:= StrPas(Z);
      P:= Pos(#9, S);
      if P <> 0 then S:= Copy(S, 1, P -1);

      W1:= Canvas.TextWidth(S);
    end;

    if R1 < W1 then R1:= W1;

    if P <> 0 then begin
      W2:= Canvas.TextWidth(Copy(Z, P +1, MaxInt));
      if R2 < W2 then R2:= W2;
    end;
  end;

  MainWidth:= R1;
  Result:= R1 + R2;

  with Options.Title do
    if (not Assigned(ParentPopupMenu98Form))
    and Visible
    and ((Align = alLeft) or (Align = alRight))
    then Inc(Result, Width);


  // fit to background
  W1:= Options.Background.Width - 2*Options.Indent -4;
  if Options.FitToBackground
  and (Result < W1)
  then Result:= W1;

end;

function TPopupMenu98Form.GetIndexAt(X, Y: Integer): Integer;
var
  I, H: Integer;
begin
  H:= Options.ItemSpace;
  Result:= -1;
  if (MenuHandle = 0)
  or (Y <= H)
  then Exit;

  with Options.Title do
    if (not Assigned(ParentPopupMenu98Form))
    and Visible
    and (((Align = alLeft) and (X <= Width))
    or   ((Align = alRight) and (X >= Buffer.Width - Width)))
    then Exit;

  mii.fMask:= miim_Type;
  for I:= 0 to MenuItemCount -1 do begin
    mii.dwTypeData:= Z;
    mii.cch:= SizeOf(Z) -1;
    if not GetMenuItemInfo(MenuHandle, I, True, mii) then Continue;

    if mii.fType and mft_Separator <> 0
    then Inc(H, Options.SeparatorHeight)
    else Inc(H, Options.ItemHeight);

    if Y <= H then begin
      Result:= I;
      Exit;
    end;

    Inc(H, Options.ItemSpace);

    if Y <= H then begin
      Result:= -1;
      Exit;
    end;
  end;
end;

function TPopupMenu98Form.CurItemTop: Integer;
var
  I, J: Integer;
begin
  J:= 0;
  Result:= 0;
  if (not Assigned(MenuItems))
  or (SelectedIndex = -1)
  or (SelectedIndex >= MenuItems.Count)
  then Exit;

  for I:= 0 to MenuItems.Count -1 do
    if MenuItems[I].Visible then
      if J = SelectedIndex
      then
        Exit
      else begin
        Inc(J);

        if MenuItems[I].Caption = '-'
        then Inc(Result, Options.SeparatorHeight)
        else Inc(Result, Options.ItemHeight);
      end;
end;

procedure TPopupMenu98Form.SetSelectedIndex(Value: Integer);
var
  Direction: Integer;
begin
  ActiveHintWindow98.ReleaseHandle;
  
  if (Value <> SelectedIndex) then begin

    if Value > FSelectedIndex then Direction:= 1
                              else Direction:= -1;

    LastSelectedIndex:= FSelectedIndex;
    FSelectedIndex:= Value;
    SelectedChanged:= True;

    if MenuHandle = 0 then Exit;

    mii.fMask:= miim_Type + miim_State;
    repeat
      // in case if LastSelectedItem = -1
      if (FSelectedIndex = LastSelectedIndex) then Break;

      if FSelectedIndex < 0 then FSelectedIndex:= MenuItemCount -1;
      if FSelectedIndex >= MenuItemCount then FSelectedIndex:= -1;

      // in other cases
      mii.dwTypeData:= Z;
      mii.cch:= SizeOf(Z) -1;
      if GetMenuItemInfo(MenuHandle, FSelectedIndex, True, mii)
      and ((mii.fType and mft_Separator = 0)
      or (FSelectedIndex = LastSelectedIndex))
      then Break;

      Inc(FSelectedIndex, Direction);

    until False;

    DrawDisabled:= True;
    Paint;

    CheckShowHint(GetCurMenuItem(False), False);
  end;
end;


{ Events }

procedure TPopupMenu98Form.WMActivate(var Msg: TWMActivate);
var
  F: Forms.TForm;
begin
  if (Msg.Active = wa_Active)
  and (Owner is Forms.TForm)
  then begin
    F:= Forms.TForm(Owner);
    while (F.FormStyle = fsMdiChild)
    and (F.Owner is Forms.TForm)
    do F:= Forms.TForm(F.Owner);

    if (F is Forms.TForm)
    and (F.Visible)
    and not (csDestroying in F.ComponentState)
    then SendMessage(F.Handle, wm_NCActivate, 1, 0);

  end
  else
    inherited;

end;

// many thanks to Jordan Russell again!...
procedure TPopupMenu98Form.WMMouseActivate(var Msg: TWMMouseActivate);
begin
  Msg.Result:= ma_NoActivate;
//  SetWindowPos(Handle, Hwnd_Top, 0, 0, 0, 0, FormFlags);
  if (Owner is Forms.TForm) then
    SetActiveWindow(Forms.TForm(Owner).Handle);
end;

procedure TPopupMenu98Form.CMMouseLeave(var Msg: TMessage);
begin
  inherited;

  if Animated then Exit;

  mii.fMask:= miim_SubMenu;
  if not (GetMenuItemInfo(MenuHandle, FSelectedIndex, True, mii)
  and (mii.hSubMenu <> 0)) then begin
    LastSelectedIndex:= FSelectedIndex;
    FSelectedIndex:= -1;
    SelectedChanged:= True;
    Paint;
  end;

  // clear status bar
  SetStatusBarText('');

  // and hint window
  ActiveHintWindow98.ReleaseHandle;
end;

procedure TPopupMenu98Form.WMSetFocus(var Msg: TWMSetFocus);
begin
  if (Animated)
  and (Owner is Forms.TForm)
  and (Forms.TForm(Owner).Visible)
  then Forms.TForm(Owner).SetFocus
  else inherited;
end;

procedure TPopupMenu98Form.CMShowingChanged(var Msg: TMessage);
begin
  // Skip Application.UpdateVisible
  // that shows annoying 'application icon' on taskbar
  if Showing
  then begin
    ShowWindow(Handle, sw_ShowNormal or sw_ShowNA);
    SetWindowPos(Handle, hwnd_Top, Left, Top, Width, Height, FormFlags);
  end
  else
    ShowWindow(Handle, sw_Hide);
end;

procedure TPopupMenu98Form.MouseDown;
begin
  ActiveHintWindow98.ReleaseHandle;
  BringMenuToFront;
end;

procedure TPopupMenu98Form.MouseUp;
var
  wID: Integer;
  MI: Menus.TMenuItem;
begin
  if Animated or KillAnimation then Exit;

  // clear status bar
  SetStatusBarText('');

  // and hint window
  ActiveHintWindow98.ReleaseHandle;

  if Assigned(CurSubMenu) then begin
    if Forms.TForm(Owner).Visible
    then Forms.TForm(Owner).SetFocus;
    Exit;
  end;

  MI:= GetCurMenuItem(False);
  if Assigned(MI) then begin

    // ordinary menu item
    if MI.Count > 0 then begin PopupCurSubMenu; Exit; end;
    if (not MI.Enabled) or (MI.Caption = '-') then Exit;

    Timer.Enabled:= False;
    KillActivePopupMenu98(True);
    KillAnimation:= True;
    SendMessage(Forms.TForm(Owner).Handle, wm_NCActivate, 1, 0);

    ShowCaret(0);
    sndPlaySound(PChar(MenuCommandSound), snd_Async + snd_NoDefault);

    if Assigned(PopupMenu.OnMenuCommand) then PopupMenu.OnMenuCommand(MI);
    if Assigned(MI.OnClick) then MI.OnClick(MI);

  end
  else begin

    mii.fMask:= miim_Type + miim_ID + miim_Submenu;
    mii.dwTypeData:= Z;
    mii.cch:= SizeOf(Z) -1;
    if (not GetMenuItemInfo(MenuHandle, SelectedIndex, True, mii))
    or (mii.fType and mft_Separator <> 0)
    or (mii.fState and mfs_Disabled <> 0)
    then Exit;

    wID:= mii.wID;
    KillActivePopupMenu98(True);
    KillAnimation:= True;
    SendMessage(Forms.TForm(Owner).Handle, wm_NCActivate, 1, 0);

    ShowCaret(0);
    sndPlaySound(PChar(MenuCommandSound), snd_Async + snd_NoDefault);

    if Assigned(PopupMenu.OnMenuCommand) then PopupMenu.OnMenuCommand(nil);
    PostMessage(Forms.TForm(Owner).Handle, wm_Command, wID, 0);

    if AssignedActiveMenu98 then
      ActiveMenu98.CheckForMdiChildMenu(nil);
  end;

end;

procedure TPopupMenu98Form.MouseMove(Shift: TShiftState; X, Y: Integer);
const
  LastX: Integer = 0;
  LastY: Integer = 0;
var
  I: Integer;
begin
  // ignore some little movements
  if Animated
  or KillAnimation
  or ((Abs(X - LastX) <= 2)
  and (Abs(Y - LastY) <= 2))
  then Exit;

  LastX:= X;
  LastY:= Y;

  // thanks to Daniel Fazekas
  // for observation of this bug
  if IgnoreFirstMouseMove then begin
    IgnoreFirstMouseMove:= False;
    Exit;
  end;

  I:= GetIndexAt(X, Y);
  if I <> SelectedIndex then begin
    Timer.Enabled:= False;

    SetStatusBarText('');
    ActiveHintWindow98.ReleaseHandle;

    LastSelectedIndex:= FSelectedIndex;
    FSelectedIndex:= I;
    SelectedChanged:= True;
    DrawDisabled:= False;
    Paint;

    CheckShowHint(GetCurMenuItem(False), False);

    Timer.Enabled:= True;
  end;
end;

// thanks for this routine to Stewart Creswell
procedure TPopupMenu98Form.SetZOrder(TopMost : Boolean);
const
  NormalWindowPos: array[Boolean] of Word = (HWND_BOTTOM, HWND_TOP);
  TopWindowPos: array[Boolean] of Integer = (HWND_NOTOPMOST, HWND_TOPMOST);
begin
  if Parent <> nil
  then
    inherited
  else
    if Handle <> 0 then
      if FormStyle = fsStayOnTop
      then SetWindowPos(Handle, TopWindowPos[TopMost], 0, 0, 0, 0, FormFlags)
      else SetWindowPos(Handle, NormalWindowPos[TopMost], 0, 0, 0, 0, FormFlags);
end;

procedure TPopupMenu98Form.PopupMenu98FormTimer(Sender: TObject);
var
  fw: HWND;
  C: TComponent;
begin
  Timer.Enabled:= False;

  if SelectedChanged then begin
    SelectedChanged:= False;
    GetCurMenuItem(False);

    // this is because DrawDisabled = True when selecton moved by kbd
    // and = False when moved by mouse :)
    if not DrawDisabled
    then CheckShowHint(CurMenuItem, True);

    if (not Assigned(CurSubMenu))
    or (CurMenuItem <> CurSubMenu.MenuItems)
    then PopupCurSubMenu;

  end;


  // close menus when owner *and* menu windows (and ActiveMdiChild) is not active
  fw:= GetForegroundWindow;
  C:= FindControl(fw);
  if not ((fw = 0)
  or (C is Forms.TForm)
  or (fw = Application.Handle)
  or (Assigned(ActiveHintWindow98) and (fw = ActiveHintWindow98.Handle)))
  then begin
    getwindowtext(fw, z, SizeOf(z) -1);

    ShowCaret(0);
    KillActivePopupMenu98(False);

    if AssignedActiveMenu98
    then ActiveMenu98.KeepSelected:= False;
  end;

  if not KillAnimation then
    Timer.Enabled:= True;
end;

procedure TPopupMenu98Form.PopupCurSubMenu;
begin
  if Assigned(CurSubMenu) then begin
    CurSubMenu.KillAnimation:= True;
    if not CurSubMenu.Animated then begin
      CurSubMenu.AnimatedHide;
      CurSubMenu.Release;
    end;
    CurSubMenu:= nil;
  end;

  ProcessPaintMessages;

  mii.fMask:= miim_State + miim_SubMenu;
  if (GetMenuItemInfo(MenuHandle, FSelectedIndex, True, mii))
  and (mii.hSubMenu <> 0)
  and (mii.fState and mfs_Disabled = 0) then try

    CurSubMenu:= TPopupMenu98Form.Create(Owner);
    CurSubMenu.Font.Assign(Font);
    CurSubMenu.ParentPopupMenu98Form:= Self;
    CurSubMenu.OpenAnimation:=  OpenAnimation;
    CurSubMenu.CloseAnimation:= CloseAnimation;
    CurSubMenu.PopupMenu:=      PopupMenu;
    CurSubMenu.MenuHandle:=     mii.hSubMenu;
    CurSubMenu.MenuItems:=      GetCurMenuItem(False);

    if (GetCurMenuItem(False) <> nil)
    and Assigned(CurMenuItem.OnClick)
    then CurMenuItem.OnClick(CurMenuItem);

    with ClientToScreen(Point(Width -7, CurItemTop)) do
      CurSubMenu.SetBounds(X, Y, 0, 0);

    CurSubMenu.Animate;

  except
    CurSubMenu:= nil
  end;
end;

procedure TPopupMenu98Form.BringMenuToFront;
var
  F: TPopupMenu98Form;
begin
  if KillAnimation then Exit;

  // process all submenus
  F:= Self;
  while Assigned(F) do begin
    SetWindowPos(F.Handle, hwnd_TopMost, F.Left, F.Top, F.Width, F.Height, FormFlags);
    F:= F.CurSubMenu;
  end;

  // hint
  if ActiveHintWindow98.Showing then
    with ActiveHintWindow98 do
      SetWindowPos(Handle, hwnd_TopMost, Left, Top, Width, Height, FormFlags);

end;

function TPopupMenu98Form.GetCurMenuItem(RaiseException: Boolean): Menus.TMenuItem;
begin
  Result:= GetMenuItemIndex(SelectedIndex, RaiseException);
end;

function TPopupMenu98Form.GetMenuItemIndex(Index: Integer; RaiseException: Boolean): Menus.TMenuItem;
var
  I, C: Integer;
begin
  Result:= nil;
  CurMenuItem:= nil;
  if not Assigned(MenuItems) then Exit;

  C:= -1;
  for I:= 0 to MenuItems.Count -1 do
    if MenuItems[I].Visible then begin
      Inc(C);
      if C = Index then begin
        CurMenuItem:= MenuItems[I];
        Result:= CurMenuItem;
        Exit;
      end;
    end;

  if RaiseException then
    raise Exception.Create('Cannot find menu item.');
end;


{ Main Routines }

function TPopupMenu98Form.Animate: Boolean;
var
  NewLeft, NewTop, NewWidth, NewHeight, MX, MY, BL: Integer;
  PlayCount, Time, TimeStart, TimeMinus: Integer;
  CurAnim: TAnimation;
  abd: TAppBarData;
  SR: TRect; // original screen rect

begin
  DX:= 1;
  DY:= 1;
  PlayCount:= 3;
  Result:= True;
  Animated:= True;
  KillAnimation:= False;
  Time:= PopupMenu.Options.Timeout;
  TimeMinus:= 0;


  // assign options property
  Options:= PopupMenu.Options;
  for BL:= 0 to PopupMenu.Submenus.Count -1 do
    if PopupMenu.Submenus[BL].MenuItem = MenuItems
    then Options:= PopupMenu.Submenus[BL];


  // calculate real coords of Windows Desktop without TaskBar
  abd.cbSize:= sizeOf(abd);
  abd.hWnd:= Handle;
  SHAppBarMessage(abm_GetTaskBarPos, abd);
  SubtractRect(SR, Rect(0, 0, Screen.Width, Screen.Height), abd.rc);

  // calculate normal height and width.
  // mx and mx -- width and height of the full opened window
  MenuItemCount:= GetMenuItemCount(MenuHandle);

  MX:= GetRealWidth + 2*Options.Indent + 10;
  MY:= GetRealHeight + 6;
  Buffer.Width:= MX;
  Buffer.Height:= MY;

  // code to arrange menu form
  BL:= 0;
  if AssignedActiveMenu98
  and (ParentPopupMenu98Form = nil)
  and (PopupMenu.Button <> nil)
  then BL:= ActiveMenu98.ClientToScreen(Point(PopupMenu.Button.Left, 0)).X;

  if (Left + MX) > SR.Right then
    if (BL <> 0)          // equals Assigned(PopupMenu.Button)
    and (PopupMenu.Button.IsVertical)
    then Left:= BL - MX
    else
      if Assigned(ParentPopupMenu98Form)
      then begin
        if OpenAnimation = anVSlide
        then Left:= ParentPopupMenu98Form.Left - MX +6
        else Left:= ParentPopupMenu98Form.Left +6;

        FromRightToLeft:= True;
      end
      else Left:= SR.Right - MX;

  if Left <= 0 then Left:= 1;

  FromBottomToTop:= not (((Top + MY) < SR.Bottom) or (Top < MY)
    or (OpenAnimation = anHSlide));

  if FromBottomToTop
  and (PopupMenu.PopupComponent is TControl)
  and (Top < TControl(PopupMenu.PopupComponent).Height)
  then FromBottomToTop:= False;

  if FromBottomToTop
  then begin
    // menu shouldn't hide a menu button
    if ((PopupMenu.PopupComponent is TToolbarButton97)
    or (PopupMenu.PopupComponent is TButton)
    or (PopupMenu.PopupComponent is TBitBtn))
    and ((PopupMenu.Button = nil) or (not PopupMenu.Button.IsVertical))
    then Top:= Top - TControl(PopupMenu.PopupComponent).Height -2;

    // ...and system menu icon too...
    if (BL = 0)
    and (PopupMenu is TSystemMenu98)
    and (PopupMenu.PopupComponent is TForm)
    then Top:= Forms.TForm(PopupMenu.PopupComponent).Top +3;
    
  end
  else
    if Top + MY > SR.Bottom
    then begin
      Top:= SR.Bottom - MY;

      if BL <> 0 then
      with PopupMenu.Button do
        if (BL <> 0)
        and (not IsVertical)
        then
          if (Self.Left + MX + Width > SR.Right)
          then Self.Left:= BL - MX
          else Self.Left:= Self.Left + Width;
    end;

  if Top <= 0 then Top:= 1;

  NewLeft:= Left;
  NewTop:= Top;
  NewWidth:= MX;
  NewHeight:= MY;

  // code for calculation animation steps
  CurAnim:= OpenAnimation;
  if CurAnim = anRandom then begin
    if Random(2) = 0 then CurAnim:= anHSlide
                     else CurAnim:= anVSlide;
  end;

  if CurAnim = anSmart then begin
    OpenAnimation:= anHSlide;
    CurAnim:= anVSlide;
  end;

  if CurAnim = anPopup then begin
    OpenAnimation:= anHSlide;
    CurAnim:= anUnfold;
  end;

  if CurAnim in [anVSlide, anUnfold] then begin
    if PopupMenu.Options.DoubleSpeed
    then DY:= MY div Round(1.2 * Options.ItemHeight) +1
    else DY:= MY div Round(0.6 * Options.ItemHeight) +1;
  end;

  if CurAnim in [anHSlide, anUnfold] then begin
    if PopupMenu.Options.DoubleSpeed
    then DX:= MX div Round(1.2 * Options.ItemHeight) +1
    else DX:= MX div Round(0.6 * Options.ItemHeight) +1;
  end;

  if DX < DY then DX:= DY;

  DY:= Round((MY div DX) * 1.5);
  DX:= Round((MX div DX) * 1.5) +1;

  if CurAnim in [anVSlide, anUnfold] then NewHeight:= 0;
  if CurAnim in [anHSlide, anUnfold] then NewWidth:=  0;

  // check if mouse already moved
  ProcessMouseMoveMessages;
  if KillAnimation or not Assigned(PopupMenu) then begin
    Animated:= False;  Result:= False;  Exit;
  end;

  // start
  PaintMenu:= True;
  Paint;
  SilentShow;

  // Animation!
  while (NewWidth < MX) or (NewHeight < MY) do begin
    // calculate the time wasted on drawing
    TimeStart:= GetCurrentTime;

    if NewWidth < MX then
      if (NewWidth + DX) >= MX then Break
      else begin
        NewWidth:= NewWidth + DX;
        if FromRightToLeft then NewLeft:= Left - DX;
      end;

    if NewHeight < MY then
      if (NewHeight + DY) >= MY then Break
      else begin
        NewHeight:= NewHeight + DY;
        if FromBottomToTop then NewTop:= Top - DY;
      end;

    SetBounds(NewLeft, NewTop, NewWidth, NewHeight);
    Paint;

    ProcessMouseMoveMessages;

    if KillAnimation or not Assigned(PopupMenu) then begin
      Animated:= False;  Result:= False;  Exit;
    end;

    ProcessPaintMessages;

    // it's necessary in case of menu already opened by another thread
    NewWidth:= Width;
    NewHeight:= Height;

    if PlayCount >=0 then
      Dec(PlayCount);

    if PlayCount = 0 then
      sndPlaySound(PChar(MenuPopupSound), snd_Async + snd_NoDefault);

    // adjust delay
    BL:= Time - (GetCurrentTime - TimeStart);
    if BL < 0 then begin
      TimeMinus:= TimeMinus + BL;
      BL:= 0;
    end;

    if TimeMinus = 0
    then
      Sleep(BL)
    else begin
      Inc(TimeMinus, BL);
      if TimeMinus > 0 then TimeMinus:= 0;
    end;

  end;

  // final setting bounds
  if FromBottomToTop then NewTop:= Top - (MY - Height) +1;
  if FromRightToLeft then NewLeft:= Left - (MX - Width);

  SetBounds(NewLeft, NewTop, MX, MY);
  Paint;

  if PlayCount > 0 then
    sndPlaySound(PChar(MenuPopupSound), snd_Async + snd_NoDefault);

  // finalization
  Animated:= False;
  KillAnimation:= False;
  Timer.Enabled:= Assigned(PopupMenu);

end;

function TPopupMenu98Form.AnimatedHide: Boolean;
var
  NewLeft, NewTop, NewWidth, NewHeight, MX, MY: Integer;
  CurAnim: TAnimation;
begin
  Timer.Enabled:= False;
  MX:= 0;
  MY:= 0;
  Result:= True;
  Animated:= True;
  KillAnimation:= False;
  NewLeft:= Left;
  NewTop:= Top;
  NewWidth:= Width;
  NewHeight:= Height;

  // event
  if Assigned(PopupMenu)
  and Assigned(PopupMenu.OnMenuClose)
  then PopupMenu.OnMenuClose(GetCurMenuItem(False));

  CurAnim:= CloseAnimation;          // current effect
  if CurAnim = anRandom then begin
    if Random(2) = 0 then CurAnim:= anHSlide
                     else CurAnim:= anVSlide;
  end;

  if CurAnim = anSmart then
    if ParentPopupMenu98Form = nil
    then CurAnim:= anVSlide
    else CurAnim:= anHSlide;

  if CurAnim = anPopup then
    if ParentPopupMenu98Form = nil
    then CurAnim:= anUnfold
    else CurAnim:= anHSlide;

  if not (CurAnim in [anVSlide, anUnfold]) then MY:= Height;
  if not (CurAnim in [anHSlide, anUnfold]) then MX:= Width;


  sndPlaySound(PChar(MenuCloseSound), snd_Async + snd_NoDefault);

  // Animation!
  while Assigned(PopupMenu) and ((Width > MX) or (Height > MY)) do begin
    if Width > MX then
      if (Width - DX) > 0
      then begin
        NewWidth:= Width - DX;
        if FromRightToLeft then NewLeft:= Left + DX;
      end
      else begin
        if FromRightToLeft then NewLeft:= Left + (MX - Width);
        NewWidth:= 0;
      end;

    if Height > MY then
      if (Height - DY) > 0
      then begin
        NewHeight:= Height - DY;
        if FromBottomToTop then NewTop:= Top + DY;
      end
      else begin
        if FromBottomToTop then NewTop:= Top + (MY - Height);
        NewHeight:= 0;
      end;

    if KillAnimation then begin
      Animated:= False;
      SilentHide;
      Exit;
    end;

    SetBounds(NewLeft, NewTop, NewWidth, NewHeight);
    Paint;
    ProcessPaintMessages;
  end;
  Animated:= False;
  KillAnimation:= False;
  SilentHide;

//  if Assigned(ParentPopupMenu98Form) then
//    with ParentPopupMenu98Form do
//      SetWindowPos(Handle, hwnd_TopMost, Left, Top, Width, Height, FormFlags);

end;

procedure TPopupMenu98Form.SilentShow;
begin
  // special check for tray-icon menu
  if Assigned(Application.MainForm)
  and not Application.MainForm.Visible
  then SetForegroundWindow(Handle);

  SetWindowPos(Handle, hwnd_TopMost, Left, Top, Width, Height, FormFlags);
  HideCaret(0);
  CheckShowHint(GetCurMenuItem(False), False);

end;

procedure TPopupMenu98Form.SilentHide;
//var
//  I: Integer;
begin
  Timer.Enabled:= False;
  if Assigned(CurSubMenu) then begin
    CurSubMenu.KillAnimation:= True;
    CurSubMenu.Release;
    CurSubMenu:= nil;
  end;

//  if Assigned(Application.MainForm)
//  and not Application.MainForm.Visible
//  then
//    Hide
//  else begin
//  ShowWindow(Handle, sw_Hide);
  SetBounds(Left, Top, 0, 0);
//  end;

  FSelectedIndex:= -1;

//  MoveControls(Self, Form1.Panel1, L);

//  for I:= 0 to ControlCount -1 do begin
//    Controls[0].Visible:= False;
//    Controls[0].Parent:= TWinControl(L[I]);
//  end;

  ProcessPaintMessages;
end;

procedure TPopupMenu98Form.Paint;
const
  ItemColors : array [Boolean] of TColor = (clMenu, clHighlight);

var
  H: Integer;
  R: TRect;
  ItemType: (itString, itSep, itShapes);

  procedure DrawTextAt(X, Y: Integer; Text: String; Default, Enabled, Selected: Boolean;
    Alignment: TAlignment);
  var
    Flags: Integer;
    IncludedBold, IncludedUnd: Boolean;
  begin
    Flags:= dt_NoClip + dt_VCenter;
    with Buffer.Canvas do begin
      IncludedBold:= Default
        and not (fsBold in Font.Style);
      IncludedUnd:= Selected
        and Enabled
        and (Options.Highlight = hiHyperlink)
        and not (fsUnderline in Font.Style);

      if IncludedBold then Font.Style:= Font.Style + [fsBold];
      if IncludedUnd  then Font.Style:= Font.Style + [fsUnderline];

      R:= Rect(X, Y, X + MainWidth, Y + H);
      Brush.Style:= bsClear;

      case Alignment of
        taRightJustify: Flags:= Flags or dt_Right;
        taCenter: Flags:= Flags or dt_Center;
      end;

      if Enabled
      then
        if Selected
        then Font.Color:= Options.Colors.HighlightText
        else Font.Color:= Options.Colors.MenuText
      else
        Font.Color:= Options.Colors.DisabledText;


      if not (Enabled or Selected) or not (Enabled or DrawDisabled) then begin
        OffsetRect(R, 1, 1);
        Font.Color:= Options.Colors.DisabledShadow;
        DrawText(Handle, PChar(Text), -1, R, Flags);
        OffsetRect(R, -1, -1);
        Font.Color:= Options.Colors.DisabledText;
      end;

      DrawText(Handle, PChar(Text), -1, R, Flags);

      if IncludedBold then Font.Style:= Font.Style - [fsBold];
      if IncludedUnd  then Font.Style:= Font.Style - [fsUnderline];
    end;
  end;

  function GetItemHeight(const M: Menus.TMenuItem): Integer;
  begin
    if M.Caption = '-' then begin
      ItemType:= itSep;
      Result:= Options.SeparatorHeight;
    end
    else
    if (M.Caption <> '') and (M.Caption[1] = ':') and (M.Caption[2] = ':') then begin
      ItemType:= itShapes;
      Result:= Options.ItemHeight;
    end
    else begin
      ItemType:= itString;
      Result:= Options.ItemHeight;
    end
  end;

var
  I, P, X, Y, menu_ItLeft, menu_ScLeft, menu_Right: Integer;
  DontFreeM, UseMenuItemBitmaps, DestroyBitmap, HighItem: Boolean;
  M: Menus.TMenuItem;
  hBmp: HBitmap;
//  DC: HDC;
//  hOld: THandle;
  Bmp: TBitmap;
  S, S1: String;
  OldPalette: HPalette;
  RestorePalette: Boolean;

begin
  if (not Assigned(MenuItems))
  and (MenuHandle = 0) then Exit;

  // initialization
  DontFreeM:= False;
  M:= nil;
  Y:= Options.ItemSpace +3;
  OldPalette:= 0;
  RestorePalette:= False;
  ItemColors[False]:= Options.Colors.Menu;
  ItemColors[True]:=  Options.Colors.Highlight;


  // bounds
  menu_ItLeft:= 0;
  menu_Right:= Buffer.Width;

  if (not Assigned(ParentPopupMenu98Form))
  and Options.Title.Visible then
    case Options.Title.Align of
      alLeft:  Inc(menu_ItLeft, Options.Title.Width);
      alRight: Dec(menu_Right,  Options.Title.Width);
    end;

  menu_ScLeft:= menu_ItLeft + MainWidth + Options.Indent +10;


  // background image
  if (not Options.Background.Empty)
  and (Options.Background.Palette <> 0)
  then begin
    OldPalette:= 0;
    RestorePalette:= False;

    if Options.Background.Palette <> 0 then begin
      OldPalette:= SelectPalette(Buffer.Canvas.Handle, Options.Background.Palette, True);
      RealizePalette(Buffer.Canvas.Handle);
      RestorePalette:= True;
    end;

  end;

  if PaintMenu
  and (Options.ItemSpace > 0)
  and not (Options.Background.Empty)
  then
    BitBlt(Buffer.Canvas.Handle, menu_ItLeft +3, 3, menu_Right - menu_ItLeft -2, Options.ItemSpace,
      Options.Background.Canvas.Handle, 0, 0, Buffer.Canvas.CopyMode);

  // paint edge
  if PaintMenu then
    with Buffer, Canvas do begin
      Font:= Self.Font;
      R:= ClipRect;
      DrawEdge(Handle, R, bdr_RaisedInner + bdr_RaisedOuter, bf_Rect);

      if Options.Title.Visible
      then Options.Title.Paint(Buffer.Canvas);

    end;

  // main drawing
  for I:= 0 to MenuItemCount -1 do begin
    // free old M
    if not DontFreeM then M.Free;

    M:= GetMenuItemIndex(I, False);
    if Assigned(M)
    then begin // standard menu item found - it's ok
      DontFreeM:= True;

      // it's still is necessary to check for menu item bitmaps
      mii.fMask:= miim_Submenu + miim_CheckMarks;
      GetMenuItemInfo(MenuHandle, I, True, mii);
    end
    else begin
      // menu item like Window Menu's item created by Windows
      // these items doesn't have corresponding TMenuItem object
      mii.dwTypeData:= Z;
      mii.cch:= SizeOf(Z) -1;
      mii.fMask:= miim_Data + miim_ID + miim_State + miim_Submenu + miim_Type + miim_CheckMarks;
      if not GetMenuItemInfo(MenuHandle, I, True, mii) then Continue;

      if mii.fType and mft_Separator <> 0 then
        Z:= '-';

      M:= NewItem(
        StrPas(Z),
        0,
        mii.fState and mfs_Checked <> 0,
        mii.fState and mfs_Disabled = 0,
        nil,
        0,
        '');

      M.Default:= mii.fState and mfs_Default <> 0;
      M.RadioItem:= mii.fType and mft_RadioCheck <> 0;

      P:= Pos(#9, M.Caption);
      if P <> 0 then begin
        M.ShortCut:= TextToShortCut(Copy(M.Caption, P +1, MaxInt));
        M.Caption:= Copy(M.Caption, 1, P -1);
      end;

      DontFreeM:= False;
    end;

    // main drawing
    with M, Buffer, Buffer.Canvas do
    if Visible then begin
      H:= GetItemHeight(M);

      if PaintMenu
      or (((I = SelectedIndex) or (I = LastSelectedIndex))
      and (SelectedIndex <> LastSelectedIndex)) then begin
        HighItem:= (I = SelectedIndex) and (ItemType = itString) and (Enabled or DrawDisabled);

        if Options.Background.Empty
        or HighItem
        then begin
          P:= 0;
          if (Options.Highlight = hiDefault)   then P:= menu_Right -3;
          if (Options.Highlight = hiRectangle) then P:= TextWidth(M.Caption) + 2*Options.Indent;

          if P <> 0 then begin
            Brush.Color:= ItemColors[HighItem];
            Brush.Style:= bsSolid;
            Windows.FillRect(Handle, Rect(menu_ItLeft +3, Y, P, Y + H), Brush.Handle);
          end;
        end
        else
          BitBlt(Canvas.Handle, menu_ItLeft +3, Y, menu_Right - menu_ItLeft -2, H + Options.ItemSpace,
            Options.Background.Canvas.Handle, 0, Y, Canvas.CopyMode);


        if ItemType = itSep then begin
          Pen.Color:= Options.Colors.Line;
          PolyLine([Point(menu_ItLeft +4, Y +3), Point(menu_Right -4, Y +3)]);
          Pen.Color:= Options.Colors.LineShadow;
          PolyLine([Point(menu_ItLeft +4, Y +4), Point(menu_Right -4, Y +4)]);
        end;

        if ItemType = itString then begin
          S:= Caption;
          P:= Pos(#9, M.Caption);
          if P <> 0 then S:= Copy(S, 1, P -1);
          DrawTextAt(menu_ItLeft + Options.Indent, Y +2, S, Default, Enabled, SelectedIndex = I, Options.Alignment);

          S1:= ShortCutToText(ShortCut);
          if (S1 = '') and (P <> 0) then S1:= Copy(M.Caption, P +1, MaxInt);
          DrawTextAt(menu_ScLeft, Y +2, S1, Default, Enabled, SelectedIndex = I, taLeftJustify);

          // submenu triangle
          if mii.hSubMenu <> 0 then begin
            if not Enabled then begin
              if SelectedIndex <> I then begin
                Brush.Color:= Options.Colors.DisabledShadow;
                Pen.Color:= Options.Colors.DisabledShadow;
                PolyGon([Point(menu_Right -13, Y +7),
                  Point(menu_Right -13, Y +13),
                  Point(menu_Right -10, Y +10)]);
              end;

              Brush.Color:= Options.Colors.DisabledText;
              Pen.Color:= Options.Colors.DisabledText;
            end
            else
              if SelectedIndex <> I then begin
                Brush.Color:= Options.Colors.MenuText;
                Pen.Color:= Options.Colors.MenuText;
              end
              else begin
                Brush.Color:= Options.Colors.HighlightText;
                Pen.Color:= Options.Colors.HighlightText;
              end;

            PolyGon([Point(menu_Right -14, Y +6),
              Point(menu_Right -14, Y +12),
              Point(menu_Right -11, Y +9)]);
          end;


          // get menu item's bitmaps
          hBmp:= 0;
          DestroyBitmap:= False;

          // default bitmap - scan resource
          if (mii.hbmpChecked = 0)
          and (mii.hbmpUnChecked = 0)
{$IFDEF NoAutoBitmaps}
          and (Copy(M.Name, 1, 6) = 'SYSTEM') // only for systemicon - if NoAutoBitmaps
{$ENDIF}
          then begin
            // first attempt - name?
            hBmp:= LoadBitmapByName(M.Name);

            // second - caption?
            // Note: I think, that this is really annoying.
            if hBmp = 0 then hBmp:= LoadBitmapByName(S);

          end;

{$IFNDEF NoAutoBitmaps}
          if (mii.hbmpChecked = 0)
          and (mii.hbmpUnChecked = 0)
          and (Copy(M.Name, 1, 5) = 'About')
          then begin
//            DC:= CreateDC('DISPLAY', nil, nil, nil);
//            hBmp:= CreateCompatibleBitmap(DC, 16, 16);
//            hOld:= SelectObject(DC, hBmp);
//            PaintMenuIcon(Forms.TForm(Owner), nil, DC);
//            SelectObject(DC, hOld);
//            DeleteDC(DC);
            Bmp:= TBitmap.Create;
            Bmp.Height:= 16;
            Bmp.Width:= 16;
            PaintMenuIcon(Forms.TForm(Owner), nil, Bmp.Canvas.Handle);
            hBmp:= Bmp.ReleaseHandle;
            Bmp.Free; 

          end;
{$ENDIF}

          if (hBmp <> 0) then begin
            if (mii.hbmpChecked = 0)   then mii.hbmpChecked:= hBmp;
            if (mii.hbmpUnChecked = 0) then mii.hbmpUnChecked:= hBmp;
            DestroyBitmap:= True;
          end;

          UseMenuItemBitmaps:=
            (Caption <> '-')
            and ((mii.hbmpChecked <> 0) or (mii.hbmpUnChecked <> 0));


          if Checked then begin
            // inner limits
            if (SelectedIndex = I) and (Enabled or DrawDisabled)
            then Pen.Color:= Options.Colors.Highlight
            else Pen.Color:= Options.Colors.Menu;

            Rectangle(menu_ItLeft + H +3, Y, menu_ItLeft + H +4, Y + H);

            R:= Rect(menu_ItLeft +3, Y, menu_ItLeft + H +3, Y + H);
            DrawEdge(Handle, R, bdr_SunkenOuter, bf_Rect);

            if Pattern = nil
            then CreatePattern;

            if (SelectedIndex = I) and (Enabled or DrawDisabled)
            then Brush.Color:= Options.Colors.Menu
            else Brush.Bitmap:= Pattern;

            FillRect(Rect(menu_ItLeft +4, Y +1, menu_ItLeft + H +2, Y + H -1));

            // draw mark
            if hBmp = 0 then
              if UseMenuItemBitmaps and (mii.hbmpChecked <> 0)
              then hBmp:= mii.hbmpChecked
              else
                if RadioItem
                then hBmp:= RadioItemBitmap
                else hBmp:= CheckMarkBitmap;

            if Enabled
            then TransBlt(Canvas, menu_ItLeft +5, Y +2, hBmp)
            else DisabledBlt(Canvas, menu_ItLeft +5, Y +2, hBmp);
            
          end
          else
            if UseMenuItemBitmaps
            and (mii.hbmpUnChecked <> 0) then begin
              if (SelectedIndex = I) and (Enabled or DrawDisabled) then begin
                Brush.Color:= Options.Colors.Menu;
                R:= Rect(menu_ItLeft +3, Y, menu_ItLeft + H +4, Y + H);
                FillRect(R);
                Dec(R.Right);
                DrawEdge(Handle, R, bdr_RaisedInner, bf_Rect);
              end;

              if Enabled
              then TransBlt(Canvas, menu_ItLeft +5, Y +2, mii.hbmpUnChecked)
              else DisabledBlt(Canvas, menu_ItLeft +5, Y +2, mii.hbmpUnChecked);

            end { draw bitmaps } ;

//            if DC <> 0 then begin
//            end;

            if DestroyBitmap then
              DeleteObject(hBmp);

        end { ItemType = itString } ;
      end;

      Inc(Y, H + Options.ItemSpace);
    end { draw next row } ;
  end;


  if PaintMenu then
    with Buffer, Canvas do begin
      // outer limits
      Brush.Style:= bsClear;
      Pen.Color:= Options.Colors.Menu;
      Polygon([Point(menu_ItLeft +2, 2), Point(menu_ItLeft +2, Buffer.Height -3),
        Point(menu_Right -3, Buffer.Height -3), Point(menu_Right -3, 2)]);


      // draw bottom background's part
      if (Y < Buffer.Height)
      and not Options.Background.Empty
      then
        BitBlt(Handle, menu_ItLeft +3, Y, menu_Right - menu_ItLeft -2, Buffer.Height - Y,
          Options.Background.Canvas.Handle, 0, Y, CopyMode);

    end;

  // background image
  if RestorePalette then
    SelectPalette(Buffer.Canvas.Handle, OldPalette, True);

  // paint childs
//  if ControlCount > 0 then
//  Form1.GraphicMenu1.PaintTo(Buffer.Canvas.Handle, 0, 0);

  // adjust buffer & draw it
  if FromBottomToTop then Y:= 0 else Y:= Buffer.Height - Height;
  if FromRightToLeft then X:= 0 else X:= Buffer.Width - Width;

  BitBlt(Canvas.Handle, 0, 0, Buffer.Width, Buffer.Height, Buffer.Canvas.Handle, X, Y, Canvas.CopyMode);

  if not DontFreeM then M.Free;
  PaintMenu:= False;
  LastSelectedIndex:= SelectedIndex;
end;


{ TMenuItem }

// a few function just for inheritation
function TMenuItem.GetBreak: TMenuBreak;
begin
  Result:= inherited Break;
end;

function TMenuItem.GetCaption: String;
begin
  Result:= inherited Caption;
end;

function TMenuItem.GetChecked: Boolean;
begin
  Result:= inherited Checked;
end;

function TMenuItem.GetDefault: Boolean;
begin
  Result:= inherited Default;
end;

function TMenuItem.GetEnabled: Boolean;
begin
  Result:= inherited Enabled;
end;

function TMenuItem.GetGroupIndex: Byte;
begin
  Result:= inherited GroupIndex;
end;

function TMenuItem.GetRadioItem: Boolean;
begin
  Result:= inherited RadioItem;
end;

function TMenuItem.GetShortCut: TShortCut;
begin
  Result:= inherited ShortCut;
end;

function TMenuItem.GetVisible: Boolean;
begin
  Result:= inherited Visible;
end;

function TMenuItem.GetHint: String;
begin
  Result:= inherited Hint;
end;

function TMenuItem.GetOnClick: TNotifyEvent;
begin
  Result:= inherited OnClick;
end;



{ routines for setting items }

const
  AM98       : TMainMenu98 = nil;
  NeedUpdate : Boolean = False;

procedure Start(MenuItem: TMenuItem);
begin
  AM98:= GetMyActiveMenu98(MenuItem);

  NeedUpdate:= True;
  if Assigned(AM98)
  then
    if AM98.Updating
    then NeedUpdate:= False
    else AM98.BeginUpdate;
end;

procedure Finish(MenuItem: TMenuItem; EasyUpdate: Boolean);
begin
  if Assigned(AM98) and NeedUpdate then
    if EasyUpdate
    then AM98.EasyUpdate(MenuItem)
    else AM98.EndUpdate;

  if ActiveMenu98 <> nil then
    HideWindowMenu(MenuItem.Owner);

  UpdateMenuItemBitmaps(TForm(MenuItem.Owner));

  AM98:= nil;
end;

procedure TMenuItem.SetBreak(Value: TMenuBreak);
begin
  Start(Self);
  inherited SetBreak(Value);
  Finish(Self, True);
end;

procedure TMenuItem.SetCaption(const Value: string);
begin
  Start(Self);
  inherited SetCaption(Value);
  Finish(Self, True);
end;

procedure TMenuItem.SetChecked(Value: Boolean);
begin
  Start(Self);
  inherited SetChecked(Value);
  Finish(Self, True);
end;

procedure TMenuItem.SetDefault(Value: Boolean);
begin
  Start(Self);
  inherited SetDefault(Value);
  Finish(Self, True);
end;

procedure TMenuItem.SetEnabled(Value: Boolean);
begin
  Start(Self);           
  inherited SetEnabled(Value);
  Finish(Self, True);
end;

procedure TMenuItem.SetGroupIndex(Value: Byte);
begin
  Start(Self);
  inherited SetGroupIndex(Value);
  Finish(Self, False);
end;

procedure TMenuItem.SetRadioItem(Value: Boolean);
begin
  Start(Self);
  inherited SetRadioItem(Value);
  Finish(Self, True);
end;

procedure TMenuItem.SetShortCut(Value: TShortCut);
begin
  Start(Self);
  inherited SetShortCut(Value);
  Finish(Self, True);
end;

procedure TMenuItem.SetVisible(Value: Boolean);
begin
  Start(Self);
  inherited SetVisible(Value);
  Finish(Self, False);
end;

procedure TMenuItem.SetHint(Value: String);
begin
  Start(Self);
  inherited Hint:= Value;
  Finish(Self, True);
end;

procedure TMenuItem.SetOnClick(Value: TNotifyEvent);
begin
  Start(Self);
  inherited OnClick:= Value;
  Finish(Self, True);
end;

function TMenuItem.IsFirstLevelItem: Boolean;
begin
  Result:= (not Assigned(Parent)) or (not Assigned(Parent.Parent));
end;

procedure TMenuItem.Insert(Index: Integer; Item: Menus.TMenuItem);
begin
  Start(Self);
  inherited Insert(Index, Item);
  Finish(Self, IsFirstLevelItem);
end;

procedure TMenuItem.Delete(Index: Integer);
begin
  Start(Self);
  inherited Delete(Index);
  Finish(Self, IsFirstLevelItem);
end;

procedure TMenuItem.Add(Item: Menus.TMenuItem);
begin
  Start(Self);
  inherited Add(Item);
  Finish(Self, IsFirstLevelItem);
end;

procedure TMenuItem.Remove(Item: Menus.TMenuItem);
begin
  Start(Self);
  inherited Remove(Item);
  Finish(Self, False);
end;



{ TForm }

procedure TForm.Loaded;
begin
  inherited;           

  if not Assigned(SystemMenu98) then
    SystemMenu98:= TSystemMenu98.Create(Self);
end;

procedure TForm.wmNCLButtonDown(var Msg: TWMNCLButtonDown);

// a few magic lines of code for one more bugfix
// thanks for these lines to Victor Santos
const
  Recall: Boolean = False;

begin
  if Recall then begin
    Recall:= False;
    Exit;
  end;

  if SystemMenu98.Enabled then begin
    if TPopupMenu98(SystemMenu98).FormOnScreen
    then KillActivePopupMenu98(False);

    if Msg.HitTest = htSysMenu then begin
      PopupSysMenu(Msg.XCursor, Msg.YCursor, True, True);
      Exit;
    end;

    if Msg.HitTest = htCaption then begin
      Recall:= True;
      PostMessage(Handle, wm_NCLButtonDown, htSysMenu, 0);
      PostMessage(Handle, wm_NCLButtonUp, htCaption, TMessage(Msg).lParam);
    end;

  end;

  inherited;
end;

procedure TForm.wmNCLButtonDblClk(var Msg: TWMNCLButtonDblClk);
begin
  if Assigned(Timer)
  then begin
    Timer.Enabled:= False;
    Timer.Free;
    Timer:= nil;
  end;

  KillActivePopupMenu98(False);

  inherited;
end;

procedure TForm.wmNCLButtonUp(var Msg: TWMNCLButtonUp);
var
  Msg1: TMsg;
begin
  if SystemMenu98.Enabled then begin
    if not PeekMessage(Msg1, 0, wm_NCMouseMove, wm_NCMouseMove, pm_NoRemove)
    then Exit;

    if (Msg.HitTest = htCaption)
    and IsIconic(Handle)
    then PopupSysMenu(Msg.XCursor, Msg.YCursor, False, False);

    if (Msg.HitTest = htSysMenu)
    or (Msg.HitTest = htCaption)
    then Exit;

  end;

  inherited;
end;

procedure TForm.wmNCRButtonDown(var Msg: TWMNCRButtonUp);
begin
  if ((Msg.HitTest <> htSysMenu)
  and (Msg.HitTest <> htCaption))
  or (not SystemMenu98.Enabled)
  then inherited;
end;

procedure TForm.wmNCRButtonUp(var Msg: TWMNCRButtonUp);
begin
  if ((Msg.HitTest = htSysMenu)
  or (Msg.HitTest = htCaption))
  and SystemMenu98.Enabled
  then PopupSysMenu(Msg.XCursor, Msg.YCursor, False, False)
  else inherited;
end;

procedure TForm.SystemClick(Sender: TObject);
begin
  PostMessage(Handle, wm_SysCommand, TComponent(Sender).Tag, 0);
end;

procedure TForm.PopupSysMenu(X, Y: Integer; DoAdjust, DoDelay: Boolean);
begin
  if AssignedActivePopupMenu98Form then Exit;

  ADoAdjust:= DoAdjust;

  if DoAdjust
  then begin
    PP:= ClientOrigin;  Dec(PP.Y);
    if WindowState = wsMinimized then
      Inc(PP.Y, 21);
  end
  else PP:= Point(X, Y);

  Timer.Free;
  Timer:= nil;

  if DoDelay then begin
    Timer:= TTimer.Create(Self);
    Timer.Interval:= 250;
    Timer.OnTimer:= DelayPopupSysMenu;
  end
  else
    DelayPopupSysMenu(nil);
    
end;

procedure TForm.DelayPopupSysMenu(Sender: TObject);
var
  SystemType: TSystemType;
  AOwner: TComponent;
begin
  Timer.Free;
  Timer:= nil;

  // determining type of the system menu 
  if Assigned(Owner)
  and (not (Owner is TApplication))
  then begin AOwner:= Owner; SystemType:= stMdiChildSystem; end
  else begin AOwner:= Self;  SystemType:= stMdiFrameSystem; end;

  // create system menu
  SystemMenu98:= TSystemMenu98(CreateSystemMenu98(AOwner, Self, SystemType, SystemMenu98));


  // PopupComponent will be assigned when
  // system menu can be adjusted by the Animate method
  if ADoAdjust
  then SystemMenu98.PopupComponent:= Self
  else SystemMenu98.PopupComponent:= nil;

  // event
  if Assigned(TSystemMenu98(SystemMenu98).OnShow)
  then TSystemMenu98(SystemMenu98).OnShow(Self);

  // adjusting menu animation type
  if SystemMenu98.AllowAdjust then
    if ADoAdjust
    then SystemMenu98.Options.OpenAnimation:= anSmart
    else SystemMenu98.Options.OpenAnimation:= anPopup;

  // displaying
  if Assigned(SystemMenu98) then
    SystemMenu98.Popup(PP.X, PP.Y);
end;



{ THintWindow98 }

constructor THintWindow98.Create(AOwner: TComponent);
begin
  inherited;

  if not Assigned(ActiveHintWindow98)
  then ActiveHintWindow98:= Self;

  Hint:= TStringList.Create;

  Timer:= TTimer.Create(Self);
  Timer.Enabled:= False;
  Timer.OnTimer:= OnTimer;

  Canvas.Font.Color:= clInfoText;
  Canvas.Brush.Color:= clInfoBk;
end;

destructor THintWindow98.Destroy;
begin
  if ActiveHintWindow98 = Self
  then ActiveHintWindow98:= nil;
  
  Timer.Free;
  Hint.Free;

  inherited;
end;

procedure THintWindow98.OnTimer(Sender: TObject);
begin
  Timer.Enabled:= False;
  ReleaseHandle;
end;

procedure THintWindow98.Paint;
var
  R: TRect;
  I, H: Integer;
begin
  H:= Canvas.TextHeight(Hint[0]);
  R:= ClientRect;

  Canvas.FillRect(R);

  Inc(R.Left, 2);
  Inc(R.Top, 2);
  R.Bottom:= H +6;

  for I:= 0 to Hint.Count -1 do begin
    DrawText(Canvas.Handle, PChar(Hint[I]), -1, R, dt_Left or dt_NoPrefix);
    OffsetRect(R, 0, H +1);
  end;
end;

procedure THintWindow98.ActivateHint(Rect: TRect; const AHint: string);
begin
  ShowHint(AHint);
end;

procedure THintWindow98.ShowHint(const AHint: string);
var
  P: TPoint;
  S, S1: String;
  MX, MY, DX, I, L: Integer;
  abd: TAppBarData;
  SR: TRect; // original screen rect
begin
  if AHint = '' then Exit;

  Color:= clInfoBk;
  Hint.Clear;

  // special symbol processing
  MX:= 0;
  MY:= 5;
  S:= Trim(GetShortHint(AHint));
  repeat
    I:= Pos('&#13;', S); L:= 4;
    if I = 0 then begin I:= Pos(#13, S); L:= 0; end;
    if I = 0 then I:= MaxInt;

    S1:= Copy(S, 1, I -1);
    Inc(MY, Canvas.TextHeight(S1) +1);
    DX:= Canvas.TextWidth(S1) +6;
    if DX > MX then MX:= DX;

    Hint.Add(S1);
    if I > Length(S) then Break;

    Delete(S, 1, I + L);

  until False;

  // getting real screen coords
  abd.cbSize:= sizeOf(abd);
  abd.hWnd:= Handle;
  SHAppBarMessage(abm_GetTaskBarPos, abd);
  SubtractRect(SR, Rect(0, 0, Screen.Width, Screen.Height), abd.rc);

  // coordinates
  GetCursorPos(P);  Inc(P.Y, 25);
  if P.X + MX > SR.Right  then P.X:= P.X - MX -10;
  if P.Y + MY > SR.Bottom then P.Y:= P.Y - MY -35;
  if P.X <= SR.Left then P.X:= SR.Left +1;
  if P.Y <= SR.Top  then P.Y:= SR.Top  +1;

  // show
  Caption:= AHint;
  BoundsRect:= Rect(0, 0, MX, MY);
  SetWindowPos(Handle, hwnd_TopMost, P.X, P.Y, 0, 0, swp_ShowWindow or swp_NoSize or swp_NoActivate);

  Timer.Interval:= Application.HintHidePause;
  Timer.Enabled:= False;
  Timer.Enabled:= True;
end;


{ TMenuItemBitmaps }

constructor TMenuItemBitmaps.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);

  FEnabled:= True;
  FCheckedBitmap:= TBitmap.Create;
  FUncheckedBitmap:= TBitmap.Create;
end;

destructor TMenuItemBitmaps.Destroy;
begin
  FCheckedBitmap.Free;
  FUncheckedBitmap.Free;

  inherited Destroy;
end;

procedure TMenuItemBitmaps.SetCheckedBitmap(Value: TBitmap);
begin
  FCheckedBitmap.Assign(Value);
  FCheckedBitmapResource:= '';

  if FSingleBitmap then
    FUncheckedBitmap.Assign(FCheckedBitmap);

  Update;
end;

procedure TMenuItemBitmaps.SetUncheckedBitmap(Value: TBitmap);
begin
  FUncheckedBitmap.Assign(Value);
  FUncheckedBitmapResource:= '';

  if FSingleBitmap then
    FCheckedBitmap.Assign(FUncheckedBitmap);

  Update;
end;

procedure TMenuItemBitmaps.SetMenuItem(Value: Menus.TMenuItem);
begin
  FMenuItem:= Value;
  Update;
end;

procedure TMenuItemBitmaps.SetSingleBitmap(Value: Boolean);
begin
  FSingleBitmap:= Value;

  if Value then begin
    if FCheckedBitmap.Empty
    then FCheckedBitmap.Assign(FUncheckedBitmap)
    else FUncheckedBitmap.Assign(FCheckedBitmap);

    Update;
  end;
end;

function GetPCharValue(Value: String): PChar;
begin
  Result:= #0;
  if Value[1] = '#' then
  try
    Result:= MakeIntResource(StrToInt(Copy(Value, 2, MaxInt)));
  except
    MessageBox(0, PChar('''' + Value + ''' is not a valid resource name'), 'Error', mb_IconError + mb_Ok);
  end
  else
    Result:= PChar(Value);
end;

procedure TMenuItemBitmaps.SetCheckedBitmapResource(Value: String);
begin
  if FCheckedBitmapResource = Value then Exit;

  FCheckedBitmapResource:= Value;
  FCheckedBitmap.ReleaseHandle;
  
  if not (csDesigning in ComponentState) then
    FCheckedBitmap.Handle:= LoadBitmap(hInstance, GetPCharValue(Value));

  Update;
end;

procedure TMenuItemBitmaps.SetUncheckedBitmapResource(Value: String);
begin
  if FUncheckedBitmapResource = Value then Exit;

  FUncheckedBitmapResource:= Value;
  FUncheckedBitmap.ReleaseHandle;

  if not (csDesigning in ComponentState) then
    FUncheckedBitmap.Handle:= LoadBitmap(hInstance, GetPCharValue(Value));

  Update;
end;

procedure TMenuItemBitmaps.Update;
begin
  if FEnabled
  and Assigned(FMenuItem)
  and Assigned(FMenuItem.Parent)
  and not (csDesigning in ComponentState) then
    SetMenuItemBitmaps(FMenuItem.Parent.Handle, FMenuItem.Command, mf_ByCommand,
      FUncheckedBitmap.Handle, FCheckedBitmap.Handle);
end;

procedure TMenuItemBitmaps.SetEnabled(Value: Boolean);
begin
  if FEnabled = Value then Exit;

  FEnabled:= Value;
  Update;

  if not FEnabled
  and Assigned(FMenuItem)
  and Assigned(FMenuItem.Parent)
  and not (csDesigning in ComponentState) then
    SetMenuItemBitmaps(FMenuItem.Parent.Handle, FMenuItem.Command, mf_ByCommand, 0, 0);
end;


{ TTitle }

constructor TTitle.Create;
begin
  inherited Create;

  FAlign:=      alLeft;
  FColorBegin:= clBlack;
  FColorEnd:=   clNavy;
  FDirection:=  True;
  FWidth:=      50;

  FFont:= TFont.Create;
  FFont.Name:= 'Arial';
  FFont.Size:=  24;
  FFont.Style:= [fsBold];
  FFont.Color:= clWhite;

  FBitmap:= TBitmap.Create;
end;

destructor TTitle.Destroy;
begin
  FFont.Free;
  FBitmap.Free;
  
  inherited;
end;

function TTitle.IsBitmapStored: Boolean;
begin
  Result:= not FBitmap.Empty;
end;

procedure TTitle.SetFont(Value: TFont);
begin
  FFont.Assign(Value);
end;

procedure TTitle.SetBitmap(Value: TBitmap);
begin
  FBitmap.Assign(Value);
end;

function TTitle.TextAlign(Width, TextWidth: Integer): Integer;
begin
  Result:= 0;
  case FAlignment of
    taLeftJustify:  Result:= Width -10;
    taCenter:       Result:= (Width + TextWidth ) div 2 +10;
    taRightJustify: Result:= TextWidth +30;
  end;
end;

procedure TTitle.Paint(Canvas: TCanvas);
var
  R: TRect;
  Bmp: TBitmap;
  lf: TLogFont;
  hfnt, holdfnt: HFont;
  X, Y, W, H, C1, C2, R1, G1, B1: Integer;
  DR, DG, DB, DH: Real;

begin
  if not FVisible then Exit;

  with Canvas.ClipRect do
    case FAlign of
      alLeft:   R:= Rect(2, 2, Width +2, Bottom -2);
      alRight:  R:= Rect(Right - Width -2, 2, Right -2, Bottom -2);
    end;

  Bmp:= TBitmap.Create;
  Bmp.Width:= R.Right - R.Left;
  Bmp.Height:= R.Bottom - R.Top;

  with Bmp.Canvas do begin
    // fill bitmap with FColorBegin and FColorEnd
    Brush.Style:= bsSolid;

    if FColorBegin <> FColorEnd then begin
      // gradient!
      DH:= Bmp.Height/256;
      C1:= ColorToRgb(FColorEnd);
      C2:= ColorToRgb(FColorBegin);
      R1:= GetRValue(C1);
      G1:= GetGValue(C1);
      B1:= GetBValue(C1);
      DR:= (GetRValue(C2) - R1 +1)/256;
      DG:= (GetGValue(C2) - G1 +1)/256;
      DB:= (GetBValue(C2) - B1 +1)/256;

      for X:= 0 to 255 do begin
        Brush.Color:= Rgb(R1 + Round(DR*X), G1 + Round(DG*X), B1 + Round(DB*X));
        FillRect(Rect(0, Round(X*DH), Bmp.Width, Round((X +1)*DH)));
      end;

    end
    else begin
      // solid color
      Brush.Color:= FColorBegin;
      FillRect(ClipRect);

    end;


    if not FBitmap.Empty then begin
      // draw bitmap
      X:= (Bmp.Width - FBitmap.Width) div 2;
      Y:= 0;

      case FAlignment of
        taLeftJustify:  Y:= Bmp.Height - FBitmap.Height;
        taRightJustify: Y:= 0;
        taCenter:       Y:= (Bmp.Height - FBitmap.Height) div 2;
      end;

      BitBlt(Bmp.Canvas.Handle, X, Y, FBitmap.Width, FBitmap.Height, FBitmap.Canvas.Handle, 0, 0, SrcCopy);
    end;

    if FText <> '' then begin
      // calculate text bounds
      Font:= Self.Font;
      W:= TextWidth(FText);
      H:= TextHeight(FText);

      // initialize TLogFont structure
      FillChar(lf, SizeOf(lf), 0);
      StrPCopy(lf.lfFaceName, Font.Name);
{$IFNDEF VER90}
      lf.lfCharSet:= Font.Charset;
{$ELSE}
      lf.lfCharSet:= Default_Charset;
{$ENDIF}
      lf.lfHeight:= Font.Height;

      if fsBold in Font.Style
      then lf.lfWeight:= fw_Bold
      else lf.lfWeight:= fw_Normal;

      lf.lfItalic:=    Integer(fsItalic    in Font.Style);
      lf.lfUnderline:= Integer(fsUnderline in Font.Style);
      lf.lfStrikeOut:= Integer(fsStrikeout in Font.Style);

      if FDirection
      then begin lf.lfEscapement:=  900;  X:= (Bmp.Width - H) div 2 -2;  Y:= TextAlign(Bmp.Height, W); end
      else begin lf.lfEscapement:= 2700;  X:= (Bmp.Width + H) div 2 +2;  Y:= Bmp.Height - TextAlign(Bmp.Height, W); end;

      hfnt:= CreateFontIndirect(lf);
      holdfnt:= SelectObject(Handle, hfnt);

      SetTextColor(Handle, ColorToRgb(Self.Font.Color));
      SetBkMode(Handle, Transparent);
      TextOut(X, Y, FText);

      SelectObject(Handle, holdfnt);
      DeleteObject(hfnt);
    end;
  end;

  BitBlt(Canvas.Handle, R.Left, R.Top, Bmp.Width, Bmp.Height, Bmp.Canvas.Handle, 0, 0, SrcCopy);
  Bmp.Free;
end;

procedure TTitle.Assign(Source: TPersistent);
begin
  if Source is TTitle then begin
    Align            := TTitle(Source).Align;
    Alignment        := TTitle(Source).Alignment;
    Text             := TTitle(Source).Text;
    ColorBegin       := TTitle(Source).ColorBegin;
    ColorEnd         := TTitle(Source).ColorEnd;
    Width            := TTitle(Source).Width;
    Direction        := TTitle(Source).Direction;
    Visible          := TTitle(Source).Visible;
    Font.Assign(TTitle(Source).Font);
    Bitmap.Assign(TTitle(Source).Bitmap);
  end
  else
    inherited;
end;


{ TColors }

constructor TColors.Create;
begin
  inherited Create;

  FBorder:=         clBtnShadow;
  FMenu:=           clMenu;
  FMenuText:=       clMenuText;
  FMenuShadow:=     clBtnShadow;
  FHighlight:=      clHighlight;
  FHighlightText:=  clHighlightText;
  FDisabledText:=   clBtnShadow;
  FDisabledShadow:= clBtnHighlight;
  FLine:=           clBtnShadow;
  FLineShadow:=     clBtnHighlight;
end;

function TColors.IsDefaultColors: Boolean;
begin
  Result:= (FBorder =      clBtnShadow)
    and (FMenu =           clMenu)
    and (FMenuText =       clMenuText)
    and (FMenuShadow =     clBtnShadow)
    and (FHighlight =      clHighlight)
    and (FHighlightText =  clHighlightText)
    and (FDisabledText =   clBtnShadow)
    and (FDisabledShadow = clBtnHighlight)
    and (FLine =           clBtnShadow)
    and (FLineShadow =     clBtnHighlight);
end;

procedure TColors.Assign(Source: TPersistent);
begin
  if Source is TColors then begin
    Border:=         TColors(Source).Border;
    Menu:=           TColors(Source).Menu;
    MenuText:=       TColors(Source).MenuText;
    Highlight:=      TColors(Source).Highlight;
    HighlightText:=  TColors(Source).HighlightText;
    DisabledText:=   TColors(Source).DisabledText;
    DisabledShadow:= TColors(Source).DisabledShadow;
    Line:=           TColors(Source).Line;
    LineShadow:=     TColors(Source).LineShadow;
  end
  else
    inherited;
end;


{ TBaseOptions }

constructor TBaseOptions.Create(Collection: TCollection);
begin
  inherited;

  FItemHeight:= 19;
  FSeparatorHeight:= 11;
  FIndent:= 27;
  FCtl3D:= True;

  FTitle:= TTitle.Create;
  FColors:= TColors.Create;
  FBackground:= TBitmap.Create;

end;

destructor TBaseOptions.Destroy;
begin
  FColors.Free;
  FTitle.Free;
  FBackground.Free;

  inherited;
end;

procedure TBaseOptions.SetBackground(Value: TBitmap);
begin
  FBackground.Assign(Value);
end;

procedure TBaseOptions.SetTitle(Value: TTitle);
begin
  FTitle.Assign(Value);
end;

procedure TBaseOptions.SetColors(Value: TColors);
begin
  FColors.Assign(Value);
end;

function TBaseOptions.IsBackgroundStored: Boolean;
begin
  Result:= not FBackground.Empty;
end;

function TBaseOptions.IsColorsStored: Boolean;
begin
  Result:= not Colors.IsDefaultColors;
end;

procedure TBaseOptions.Assign(Source: TPersistent);
begin
  if Source is TBaseOptions then begin
    FItemHeight:=      TBaseOptions(Source).FItemHeight;
    FSeparatorHeight:= TBaseOptions(Source).FSeparatorHeight;
    FIndent:=          TBaseOptions(Source).FIndent;
    FTextAlignment:=   TBaseOptions(Source).FTextAlignment;
    FHighlight:=       TBaseOptions(Source).FHighlight;
    FCtl3D:=           TBaseOptions(Source).FCtl3D;
    FFitToBackground:= TBaseOptions(Source).FFitToBackground;
    FItemSpace:=       TBaseOptions(Source).FItemSpace;
    FTitle.Assign(TBaseOptions(Source).FTitle);
    FBackground.Assign(TBaseOptions(Source).FBackground);
    FColors.Assign(TBaseOptions(Source).FColors);
  end
  else
    inherited;
end;


{ TOptions }

constructor TOptions.Create(Collection: TCollection);
begin
  inherited;

  FTimeout:=        20;
  FDoubleSpeed:=    False;
  FOpenAnimation:=  anSmart;
  FCloseAnimation:= anSmart;

end;

procedure TOptions.Assign(Source: TPersistent);
begin
  inherited;

  if Source is TOptions then begin
    FTimeout:=        TOptions(Source).FTimeout;
    FDoubleSpeed:=    TOptions(Source).FDoubleSpeed;
    FOpenAnimation:=  TOptions(Source).FOpenAnimation;
    FCloseAnimation:= TOptions(Source).FCloseAnimation;
  end;
end;


{ TSubmenuOptions }

constructor TSubmenuOptions.Create(Collection: TCollection);
begin
  inherited;
  FParentFont:= True;
  FFont:= TFont.Create;
end;

destructor TSubmenuOptions.Destroy;
begin
  FFont.Free;
  inherited;
end;

procedure TSubmenuOptions.SetFont(Value: TFont);
var
  H: Integer;
begin
  FFont.Assign(Value);

  H:= - FFont.Height + 8;
  if FItemHeight < H then FItemHeight:= H;
end;

procedure TSubmenuOptions.SetParentFont(Value: Boolean);
begin
  FParentFont:= Value;
end;

function TSubmenuOptions.IsFontStored: Boolean;
begin
  Result:= not FParentFont;
end;

procedure TSubmenuOptions.Assign(Source: TPersistent);
begin
  inherited;

  if Source is TSubmenuOptions then begin
    FFont.Assign(TSubmenuOptions(Source).FFont);
    FParentFont:=    TSubmenuOptions(Source).FParentFont;
    FShowHint:=      TSubmenuOptions(Source).FShowHint;
    FMenuItem:=      TSubmenuOptions(Source).FMenuItem;
  end;
end;

{$IFDEF VER100}
function TSubmenuOptions.GetDisplayName: String;
begin
  try
    if Assigned(FMenuItem) then begin
      Result:= RemoveAmpersands(FMenuItem.Caption);
      Exit;
    end;
  except
  end;

  FMenuItem:= nil;
  Result:= inherited GetDisplayName;
end;
{$ENDIF}

procedure TSubmenuOptions.SetMenuItem(Value: Menus.TMenuItem);
begin
  if FMenuItem = Value then Exit;

  if Value.Count = 0 then
    raise Exception.Create(Value.Name + ' doesn''t contains any submenu items');

  FMenuItem:= Value;
end;


{ TSubmenus }

constructor TSubmenus.Create(Persistent: TPersistent);
begin
  inherited Create(TSubmenuOptions);
  FPersistent:= Persistent;
end;

function TSubmenus.Add: TSubmenuOptions;
begin
  Result:= TSubmenuOptions(inherited Add);
end;

function TSubmenus.GetItem(Index: Integer): TSubmenuOptions;
begin
  Result:= TSubmenuOptions(inherited GetItem(Index));
end;

{$IFDEF VER100}
function TSubmenus.GetOwner: TPersistent;
begin
  Result:= FPersistent;
end;
{$ENDIF}

procedure TSubmenus.SetItem(Index: Integer; Value: TSubmenuOptions);
begin
  inherited SetItem(Index, Value);
end;

procedure TSubmenus.Update(Item: TCollectionItem);
begin
end;



initialization

  // randomization
  Randomize;

  // hint window
  ActiveHintWindow98:= THintWindow98.Create(nil);

  // active menu98 list for multiforms
  ActiveMenu98List:= TList.Create;

  // structure for quering menus
  mii.cbSize:= SizeOf(mii);

  // check marks
  CheckMarkBitmap:= LoadBitmap(HInstance, 'CHECKMARK');
  RadioItemBitmap:= LoadBitmap(HInstance, 'RADIOITEM');

  // locking
  DesktopWindow:= GetDesktopWindow;


finalization

  if HGetMessageHook <> 0 then begin
    UnhookWindowsHookEx(HGetMessageHook);
    HGetMessageHook:= 0;
  end;

  Pattern.Free;

  ActiveHintWindow98.Free;

  ActiveMenu98List.Free;

  DeleteObject(CheckMarkBitmap);
  DeleteObject(RadioItemBitmap);

end.


